"""
    :codeauthor: Rupesh Tare <rupesht@saltstack.com>
"""


import os.path

import salt.modules.linux_lvm as linux_lvm
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class LinuxLVMTestCase(TestCase, LoaderModuleMockMixin):
    """
    TestCase for the salt.modules.linux_lvm module
    """

    def setup_loader_modules(self):
        return {linux_lvm: {}}

    def test_version(self):
        """
        Tests LVM version info from lvm version
        """
        mock = MagicMock(
            return_value=(
                "  LVM version:     2.02.168(2) (2016-11-30)\n"
                "  Library version: 1.03.01 (2016-11-30)\n"
                "  Driver version:  4.35.0\n"
            )
        )
        with patch.dict(linux_lvm.__salt__, {"cmd.run": mock}):
            self.assertEqual(linux_lvm.version(), "2.02.168(2) (2016-11-30)")

    def test_fullversion(self):
        """
        Tests all version info from lvm version
        """
        mock = MagicMock(
            return_value=(
                "  LVM version:     2.02.168(2) (2016-11-30)\n"
                "  Library version: 1.03.01 (2016-11-30)\n"
                "  Driver version:  4.35.0\n"
            )
        )
        with patch.dict(linux_lvm.__salt__, {"cmd.run": mock}):
            self.assertDictEqual(
                linux_lvm.fullversion(),
                {
                    "LVM version": "2.02.168(2) (2016-11-30)",
                    "Library version": "1.03.01 (2016-11-30)",
                    "Driver version": "4.35.0",
                },
            )

    def test_pvdisplay(self):
        """
        Tests information about the physical volume(s)
        """
        mock = MagicMock(return_value={"retcode": 1})
        with patch.dict(linux_lvm.__salt__, {"cmd.run_all": mock}):
            self.assertDictEqual(linux_lvm.pvdisplay(), {})

        mock = MagicMock(return_value={"retcode": 1})
        with patch.dict(linux_lvm.__salt__, {"cmd.run_all": mock}):
            self.assertDictEqual(linux_lvm.pvdisplay(quiet=True), {})
            mock.assert_called_with(
                ["pvdisplay", "-c"], ignore_retcode=True, python_shell=False
            )

        mock = MagicMock(return_value={"retcode": 0, "stdout": "A:B:C:D:E:F:G:H:I:J:K"})
        with patch.dict(linux_lvm.__salt__, {"cmd.run_all": mock}):
            self.assertDictEqual(
                linux_lvm.pvdisplay(),
                {
                    "A": {
                        "Allocated Physical Extents": "K",
                        "Current Logical Volumes Here": "G",
                        "Free Physical Extents": "J",
                        "Internal Physical Volume Number": "D",
                        "Physical Extent Size (kB)": "H",
                        "Physical Volume (not) Allocatable": "F",
                        "Physical Volume Device": "A",
                        "Physical Volume Size (kB)": "C",
                        "Physical Volume Status": "E",
                        "Total Physical Extents": "I",
                        "Volume Group Name": "B",
                    }
                },
            )

            mockpath = MagicMock(return_value="Z")
            with patch.object(os.path, "realpath", mockpath):
                self.assertDictEqual(
                    linux_lvm.pvdisplay(real=True),
                    {
                        "Z": {
                            "Allocated Physical Extents": "K",
                            "Current Logical Volumes Here": "G",
                            "Free Physical Extents": "J",
                            "Internal Physical Volume Number": "D",
                            "Physical Extent Size (kB)": "H",
                            "Physical Volume (not) Allocatable": "F",
                            "Physical Volume Device": "A",
                            "Physical Volume Size (kB)": "C",
                            "Physical Volume Status": "E",
                            "Real Physical Volume Device": "Z",
                            "Total Physical Extents": "I",
                            "Volume Group Name": "B",
                        }
                    },
                )

    def test_vgdisplay(self):
        """
        Tests information about the volume group(s)
        """
        mock = MagicMock(return_value={"retcode": 1})
        with patch.dict(linux_lvm.__salt__, {"cmd.run_all": mock}):
            self.assertDictEqual(linux_lvm.vgdisplay(), {})

        mock = MagicMock(return_value={"retcode": 1})
        with patch.dict(linux_lvm.__salt__, {"cmd.run_all": mock}):
            self.assertDictEqual(linux_lvm.vgdisplay(quiet=True), {})
            mock.assert_called_with(
                ["vgdisplay", "-c"], ignore_retcode=True, python_shell=False
            )

        mock = MagicMock(
            return_value={"retcode": 0, "stdout": "A:B:C:D:E:F:G:H:I:J:K:L:M:N:O:P:Q"}
        )
        with patch.dict(linux_lvm.__salt__, {"cmd.run_all": mock}):
            self.assertDictEqual(
                linux_lvm.vgdisplay(),
                {
                    "A": {
                        "Actual Physical Volumes": "K",
                        "Allocated Physical Extents": "O",
                        "Current Logical Volumes": "F",
                        "Current Physical Volumes": "J",
                        "Free Physical Extents": "P",
                        "Internal Volume Group Number": "D",
                        "Maximum Logical Volume Size": "H",
                        "Maximum Logical Volumes": "E",
                        "Maximum Physical Volumes": "I",
                        "Open Logical Volumes": "G",
                        "Physical Extent Size (kB)": "M",
                        "Total Physical Extents": "N",
                        "UUID": "Q",
                        "Volume Group Access": "B",
                        "Volume Group Name": "A",
                        "Volume Group Size (kB)": "L",
                        "Volume Group Status": "C",
                    }
                },
            )

    def test_lvdisplay(self):
        """
        Return information about the logical volume(s)
        """
        mock = MagicMock(return_value={"retcode": 1})
        with patch.dict(linux_lvm.__salt__, {"cmd.run_all": mock}):
            self.assertDictEqual(linux_lvm.lvdisplay(), {})

        mock = MagicMock(
            return_value={"retcode": 0, "stdout": "A:B:C:D:E:F:G:H:I:J:K:L:M"}
        )
        with patch.dict(linux_lvm.__salt__, {"cmd.run_all": mock}):
            self.assertDictEqual(
                linux_lvm.lvdisplay(),
                {
                    "A": {
                        "Allocated Logical Extents": "I",
                        "Allocation Policy": "J",
                        "Current Logical Extents Associated": "H",
                        "Internal Logical Volume Number": "E",
                        "Logical Volume Access": "C",
                        "Logical Volume Name": "A",
                        "Logical Volume Size": "G",
                        "Logical Volume Status": "D",
                        "Major Device Number": "L",
                        "Minor Device Number": "M",
                        "Open Logical Volumes": "F",
                        "Read Ahead Sectors": "K",
                        "Volume Group Name": "B",
                    }
                },
            )

    def test_pvcreate(self):
        """
        Tests for set a physical device to be used as an LVM physical volume
        """
        self.assertEqual(
            linux_lvm.pvcreate(""), "Error: at least one device is required"
        )

        self.assertEqual(linux_lvm.pvcreate("A"), "A does not exist")

        # pvdisplay() would be called by pvcreate() twice: firstly to check
        # whether a device is already initialized for use by LVM and then to
        # ensure that the pvcreate executable did its job correctly.
        pvdisplay = MagicMock(side_effect=[False, True])
        with patch("salt.modules.linux_lvm.pvdisplay", pvdisplay):
            with patch.object(os.path, "exists", return_value=True):
                ret = {
                    "stdout": "saltines",
                    "stderr": "cheese",
                    "retcode": 0,
                    "pid": "1337",
                }
                mock = MagicMock(return_value=ret)
                with patch.dict(linux_lvm.__salt__, {"cmd.run_all": mock}):
                    self.assertEqual(linux_lvm.pvcreate("A", metadatasize=1000), True)

    def test_pvcreate_existing_pvs(self):
        """
        Test a scenario when all the submitted devices are already LVM PVs.
        """
        pvdisplay = MagicMock(return_value=True)
        with patch("salt.modules.linux_lvm.pvdisplay", pvdisplay):
            with patch.object(os.path, "exists", return_value=True):
                ret = {
                    "stdout": "saltines",
                    "stderr": "cheese",
                    "retcode": 0,
                    "pid": "1337",
                }
                cmd_mock = MagicMock(return_value=ret)
                with patch.dict(linux_lvm.__salt__, {"cmd.run_all": cmd_mock}):
                    self.assertEqual(linux_lvm.pvcreate("A", metadatasize=1000), True)
                    self.assertTrue(cmd_mock.call_count == 0)

    def test_pvremove_not_pv(self):
        """
        Tests for remove a physical device not being used as an LVM physical volume
        """
        pvdisplay = MagicMock(return_value=False)
        with patch("salt.modules.linux_lvm.pvdisplay", pvdisplay):
            self.assertEqual(
                linux_lvm.pvremove("A", override=False), "A is not a physical volume"
            )

        pvdisplay = MagicMock(return_value=False)
        with patch("salt.modules.linux_lvm.pvdisplay", pvdisplay):
            self.assertEqual(linux_lvm.pvremove("A"), True)

    def test_pvremove(self):
        """
        Tests for remove a physical device being used as an LVM physical volume
        """
        pvdisplay = MagicMock(return_value=False)
        with patch("salt.modules.linux_lvm.pvdisplay", pvdisplay):
            mock = MagicMock(return_value=True)
            with patch.dict(linux_lvm.__salt__, {"lvm.pvdisplay": mock}):
                ret = {
                    "stdout": "saltines",
                    "stderr": "cheese",
                    "retcode": 0,
                    "pid": "1337",
                }
                mock = MagicMock(return_value=ret)
                with patch.dict(linux_lvm.__salt__, {"cmd.run_all": mock}):
                    self.assertEqual(linux_lvm.pvremove("A"), True)

    def test_pvresize_not_pv(self):
        """
        Tests for resize a physical device not being used as an LVM physical volume
        """
        pvdisplay = MagicMock(return_value=False)
        with patch("salt.modules.linux_lvm.pvdisplay", pvdisplay):
            self.assertEqual(
                linux_lvm.pvresize("A", override=False), "A is not a physical volume"
            )

        pvdisplay = MagicMock(return_value=False)
        with patch("salt.modules.linux_lvm.pvdisplay", pvdisplay):
            self.assertEqual(linux_lvm.pvresize("A"), True)

    def test_pvresize(self):
        """
        Tests for resize a physical device being used as an LVM physical volume
        """
        pvdisplay = MagicMock(return_value=False)
        with patch("salt.modules.linux_lvm.pvdisplay", pvdisplay):
            mock = MagicMock(return_value=True)
            with patch.dict(linux_lvm.__salt__, {"lvm.pvdisplay": mock}):
                ret = {
                    "stdout": "saltines",
                    "stderr": "cheese",
                    "retcode": 0,
                    "pid": "1337",
                }
                mock = MagicMock(return_value=ret)
                with patch.dict(linux_lvm.__salt__, {"cmd.run_all": mock}):
                    self.assertEqual(linux_lvm.pvresize("A"), True)

    def test_vgcreate(self):
        """
        Tests create an LVM volume group
        """
        self.assertEqual(
            linux_lvm.vgcreate("", ""), "Error: vgname and device(s) are both required"
        )

        mock = MagicMock(return_value={"retcode": 0, "stderr": ""})
        with patch.dict(linux_lvm.__salt__, {"cmd.run_all": mock}):
            with patch.object(linux_lvm, "vgdisplay", return_value={}):
                self.assertDictEqual(
                    linux_lvm.vgcreate("fakevg", "B"),
                    {
                        "Output from vgcreate": (
                            'Volume group "fakevg" successfully created'
                        )
                    },
                )

    def test_vgextend(self):
        """
        Tests add physical volumes to an LVM volume group
        """
        self.assertEqual(
            linux_lvm.vgextend("", ""), "Error: vgname and device(s) are both required"
        )

        mock = MagicMock(return_value={"retcode": 0, "stderr": ""})
        with patch.dict(linux_lvm.__salt__, {"cmd.run_all": mock}):
            with patch.object(linux_lvm, "vgdisplay", return_value={}):
                self.assertDictEqual(
                    linux_lvm.vgextend("fakevg", "B"),
                    {
                        "Output from vgextend": (
                            'Volume group "fakevg" successfully extended'
                        )
                    },
                )

    def test_lvcreate(self):
        """
        Test create a new logical volume, with option
        for which physical volume to be used
        """
        self.assertEqual(
            linux_lvm.lvcreate(None, None, 1, 1),
            "Error: Please specify only one of size or extents",
        )

        self.assertEqual(
            linux_lvm.lvcreate(None, None, None, None),
            "Error: Either size or extents must be specified",
        )

        self.assertEqual(
            linux_lvm.lvcreate(None, None, thinvolume=True, thinpool=True),
            "Error: Please set only one of thinvolume or thinpool to True",
        )

        self.assertEqual(
            linux_lvm.lvcreate(None, None, thinvolume=True, extents=1),
            "Error: Thin volume size cannot be specified as extents",
        )

        mock = MagicMock(return_value={"retcode": 0, "stderr": ""})
        with patch.dict(linux_lvm.__salt__, {"cmd.run_all": mock}):
            with patch.object(linux_lvm, "lvdisplay", return_value={}):
                self.assertDictEqual(
                    linux_lvm.lvcreate(None, None, None, 1),
                    {"Output from lvcreate": 'Logical volume "None" created.'},
                )

    def test_lvcreate_with_force(self):
        """
        Test create a new logical volume, with option
        for which physical volume to be used
        """
        mock = MagicMock(return_value={"retcode": 0, "stderr": ""})
        with patch.dict(linux_lvm.__salt__, {"cmd.run_all": mock}):
            with patch.object(linux_lvm, "lvdisplay", return_value={}):
                self.assertDictEqual(
                    linux_lvm.lvcreate(None, None, None, 1, force=True),
                    {"Output from lvcreate": 'Logical volume "None" created.'},
                )

    def test_lvcreate_extra_arguments_no_parameter(self):
        extra_args = {
            "nosync": None,
            "noudevsync": None,
            "ignoremonitoring": None,
            "thin": None,
        }
        mock = MagicMock(return_value={"retcode": 0, "stderr": ""})
        with patch.dict(linux_lvm.__salt__, {"cmd.run_all": mock}):
            with patch.object(linux_lvm, "lvdisplay", return_value={}):
                self.assertDictEqual(
                    linux_lvm.lvcreate(None, None, None, 1, **extra_args),
                    {"Output from lvcreate": 'Logical volume "None" created.'},
                )
        expected_args = ["--{}".format(arg) for arg in extra_args]
        processed_extra_args = mock.call_args.args[0][-(len(extra_args) + 1) : -1]
        self.assertTrue(all([arg in expected_args for arg in processed_extra_args]))

    def test_lvcreate_invalid_extra_parameter(self):
        invalid_parameter = {"foo": "bar"}
        mock = MagicMock(return_value={"retcode": 0, "stderr": ""})
        with patch.dict(linux_lvm.__salt__, {"cmd.run_all": mock}):
            with patch.object(linux_lvm, "lvdisplay", return_value={}):
                self.assertDictEqual(
                    linux_lvm.lvcreate(None, None, None, 1, **invalid_parameter),
                    {"Output from lvcreate": 'Logical volume "None" created.'},
                )
        processed_command = mock.call_args.args[0]
        self.assertFalse("--foo" in processed_command)

    def test_vgremove(self):
        """
        Tests to remove an LVM volume group
        """
        mock = MagicMock(
            return_value={
                "retcode": 0,
                "stdout": '  Volume group "fakevg" successfully removed',
            }
        )
        with patch.dict(linux_lvm.__salt__, {"cmd.run_all": mock}):
            self.assertEqual(
                linux_lvm.vgremove("fakevg"),
                'Volume group "fakevg" successfully removed',
            )

    def test_lvremove(self):
        """
        Test to remove a given existing logical volume
        from a named existing volume group
        """
        mock = MagicMock(
            return_value={
                "retcode": 0,
                "stdout": '  Logical volume "lvtest" successfully removed',
            }
        )
        with patch.dict(linux_lvm.__salt__, {"cmd.run_all": mock}):
            self.assertEqual(
                linux_lvm.lvremove("fakelv", "fakevg"),
                'Logical volume "fakelv" successfully removed',
            )

    def test_lvresize(self):
        """
        Tests to resize an LVM logical volume
        """
        self.assertEqual(linux_lvm.lvresize(1, None, 1), {})

        self.assertEqual(linux_lvm.lvresize(None, None, None), {})

        mock = MagicMock(return_value={"retcode": 0, "stderr": ""})
        with patch.dict(linux_lvm.__salt__, {"cmd.run_all": mock}):
            self.assertDictEqual(
                linux_lvm.lvresize(12, "/dev/fakevg/fakelv"),
                {
                    "Output from lvresize": (
                        'Logical volume "/dev/fakevg/fakelv" successfully resized.'
                    )
                },
            )

    def test_lvextend(self):
        """
        Tests to extend an LVM logical volume
        """
        self.assertEqual(linux_lvm.lvextend(1, None, 1), {})

        self.assertEqual(linux_lvm.lvextend(None, None, None), {})

        mock = MagicMock(return_value={"retcode": 0, "stderr": ""})
        with patch.dict(linux_lvm.__salt__, {"cmd.run_all": mock}):
            self.assertDictEqual(
                linux_lvm.lvextend(12, "/dev/fakevg/fakelv"),
                {
                    "Output from lvextend": (
                        'Logical volume "/dev/fakevg/fakelv" successfully extended.'
                    )
                },
            )



"""
Test for the chocolatey module
"""


import os

import salt.modules.chocolatey as chocolatey
import salt.utils
import salt.utils.platform
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase, skipIf


@skipIf(not salt.utils.platform.is_windows(), "Not a Windows system")
class ChocolateyTestCase(TestCase, LoaderModuleMockMixin):
    """
    Chocolatey private functions tests
    """

    @classmethod
    def setUpClass(cls):
        cls.choco_path = "C:\\path\\to\\chocolatey.exe"
        cls.choco_path_pd = os.path.join(
            os.environ.get("ProgramData"), "Chocolatey", "bin", "chocolatey.exe"
        )
        cls.choco_path_sd = os.path.join(
            os.environ.get("SystemDrive"), "Chocolatey", "bin", "chocolatey.bat"
        )
        cls.mock_false = MagicMock(return_value=False)
        cls.mock_true = MagicMock(return_value=True)

    @classmethod
    def tearDownClass(cls):
        del cls.choco_path
        del cls.choco_path_pd
        del cls.choco_path_sd
        del cls.mock_false
        del cls.mock_true

    def setup_loader_modules(self):
        return {chocolatey: {"__context__": {}, "__salt__": {}}}

    def test__clear_context(self):
        """
        Tests _clear_context function
        """
        context = {
            "chocolatey._yes": ["--yes"],
            "chocolatey._path": self.choco_path,
            "chocolatey._version": "0.9.9",
        }
        with patch.dict(chocolatey.__context__, context):
            chocolatey._clear_context()
            # Did it clear all chocolatey items from __context__P?
            self.assertEqual(chocolatey.__context__, {})

    def test__yes_context(self):
        """
        Tests _yes function when it exists in __context__
        """
        with patch.dict(chocolatey.__context__, {"chocolatey._yes": ["--yes"]}):
            result = chocolatey._yes()
            expected = ["--yes"]
            # Did it return correctly
            self.assertListEqual(result, expected)
            # Did it populate __context__
            self.assertEqual(chocolatey.__context__["chocolatey._yes"], expected)

    def test__yes_version_greater(self):
        """
        Test _yes when Chocolatey version is greater than 0.9.9
        """
        mock_version = MagicMock(return_value="10.0.0")
        with patch("salt.modules.chocolatey.chocolatey_version", mock_version):
            result = chocolatey._yes()
            expected = ["--yes"]
            # Did it return correctly
            self.assertListEqual(result, expected)
            # Did it populate __context__
            self.assertEqual(chocolatey.__context__["chocolatey._yes"], expected)

    def test__yes_version_less_than(self):
        """
        Test _yes when Chocolatey version is less than 0.9.9
        """
        mock_version = MagicMock(return_value="0.9.0")
        with patch("salt.modules.chocolatey.chocolatey_version", mock_version):
            result = chocolatey._yes()
            expected = []
            # Did it return correctly
            self.assertListEqual(result, expected)
            # Did it populate __context__
            self.assertEqual(chocolatey.__context__["chocolatey._yes"], expected)

    def test__find_chocolatey_context(self):
        """
        Test _find_chocolatey when it exists in __context__
        """
        with patch.dict(chocolatey.__context__, {"chocolatey._path": self.choco_path}):
            result = chocolatey._find_chocolatey()
            expected = self.choco_path
            self.assertEqual(result, expected)

    def test__find_chocolatey_which(self):
        """
        Test _find_chocolatey when found with `cmd.which`
        """
        mock_which = MagicMock(return_value=self.choco_path)
        with patch.dict(chocolatey.__salt__, {"cmd.which": mock_which}):
            result = chocolatey._find_chocolatey()
            expected = self.choco_path
            # Does it return the correct path
            self.assertEqual(result, expected)
            # Does it populate __context__
            self.assertEqual(chocolatey.__context__["chocolatey._path"], expected)

    def test__find_chocolatey_programdata(self):
        """
        Test _find_chocolatey when found in ProgramData
        """
        with patch.dict(chocolatey.__salt__, {"cmd.which": self.mock_false}), patch(
            "os.path.isfile", self.mock_true
        ):
            result = chocolatey._find_chocolatey()
            expected = self.choco_path_pd
            # Does it return the correct path
            self.assertEqual(result, expected)
            # Does it populate __context__
            self.assertEqual(chocolatey.__context__["chocolatey._path"], expected)

    def test__find_chocolatey_systemdrive(self):
        """
        Test _find_chocolatey when found on SystemDrive (older versions)
        """
        with patch.dict(chocolatey.__salt__, {"cmd.which": self.mock_false}), patch(
            "os.path.isfile", MagicMock(side_effect=[False, True])
        ):
            result = chocolatey._find_chocolatey()
            expected = self.choco_path_sd
            # Does it return the correct path
            self.assertEqual(result, expected)
            # Does it populate __context__
            self.assertEqual(chocolatey.__context__["chocolatey._path"], expected)

    def test_version_check_remote_false(self):
        """
        Test version when remote is False
        """
        list_return_value = {"ack": ["3.1.1"]}
        with patch.object(chocolatey, "list_", return_value=list_return_value):
            expected = {"ack": ["3.1.1"]}
            result = chocolatey.version("ack", check_remote=False)
            self.assertEqual(result, expected)

    def test_version_check_remote_true(self):
        """
        Test version when remote is True
        """
        list_side_effect = [
            {"ack": ["3.1.1"]},
            {"ack": ["3.1.1"], "Wolfpack": ["3.0.17"], "blackbird": ["1.0.79.3"]},
        ]
        with patch.object(chocolatey, "list_", side_effect=list_side_effect):
            expected = {"ack": {"available": ["3.1.1"], "installed": ["3.1.1"]}}
            result = chocolatey.version("ack", check_remote=True)
            self.assertEqual(result, expected)

    def test_version_check_remote_true_not_available(self):
        """
        Test version when remote is True but remote version is unavailable
        """
        list_side_effect = [
            {"ack": ["3.1.1"]},
            {"Wolfpack": ["3.0.17"], "blackbird": ["1.0.79.3"]},
        ]
        with patch.object(chocolatey, "list_", side_effect=list_side_effect):
            expected = {"ack": {"installed": ["3.1.1"]}}
            result = chocolatey.version("ack", check_remote=True)
            self.assertEqual(result, expected)



import logging

import pytest
import salt.states.file as filestate

log = logging.getLogger(__name__)


@pytest.mark.skip_on_windows(reason="Do not run on Windows")
def test__find_keep_files_unix():
    keep = filestate._find_keep_files(
        "/test/parent_folder", ["/test/parent_folder/meh.txt"]
    )
    expected = [
        "/",
        "/test",
        "/test/parent_folder",
        "/test/parent_folder/meh.txt",
    ]
    actual = sorted(list(keep))
    assert actual == expected, actual


@pytest.mark.skip_unless_on_windows(reason="Do not run except on Windows")
def test__find_keep_files_win32():
    """
    Test _find_keep_files. The `_find_keep_files` function is only called by
    _clean_dir.
    """
    keep = filestate._find_keep_files(
        "c:\\test\\parent_folder",
        ["C:\\test\\parent_folder\\meh-1.txt", "C:\\Test\\Parent_folder\\Meh-2.txt"],
    )
    expected = [
        "c:\\",
        "c:\\test",
        "c:\\test\\parent_folder",
        "c:\\test\\parent_folder\\meh-1.txt",
        "c:\\test\\parent_folder\\meh-2.txt",
    ]
    actual = sorted(list(keep))
    assert actual == expected


@pytest.mark.skip_unless_on_windows(reason="Do not run except on Windows")
def test__clean_dir_win32():
    """
    Test _clean_dir to ensure that regardless of case, we keep all files
    requested and do not delete any. Therefore, the expected list should
    be empty for this test.
    """
    keep = filestate._clean_dir(
        "c:\\test\\parent_folder",
        [
            "C:\\test\\parent_folder\\meh-1.txt",
            "C:\\Test\\Parent_folder\\Meh-2.txt",
        ],
        exclude_pat=None,
    )
    actual = sorted(list(keep))
    expected = []
    assert actual == expected


@pytest.mark.skip_unless_on_darwin(reason="Do not run except on OS X")
def test__find_keep_files_darwin():
    """
    Test _clean_dir to ensure that regardless of case, we keep all files
    requested and do not delete any. Therefore, the expected list should
    be empty for this test.
    """
    keep = filestate._clean_dir(
        "/test/parent_folder",
        [
            "/test/folder/parent_folder/meh-1.txt",
            "/Test/folder/Parent_Folder/Meh-2.txt",
        ],
        exclude_pat=None,
    )
    actual = sorted(list(keep))
    expected = []
    assert actual == expected



import logging
from collections import namedtuple

import salt.utils.args
from salt.exceptions import SaltInvocationError
from tests.support.mock import DEFAULT, patch
from tests.support.unit import TestCase

log = logging.getLogger(__name__)


class ArgsTestCase(TestCase):
    """
    TestCase for salt.utils.args module
    """

    def test_condition_input_string(self):
        """
        Test passing a jid on the command line
        """
        cmd = salt.utils.args.condition_input(
            ["*", "foo.bar", 20141020201325675584], None
        )
        self.assertIsInstance(cmd[2], str)

    def test_clean_kwargs(self):
        self.assertDictEqual(salt.utils.args.clean_kwargs(foo="bar"), {"foo": "bar"})
        self.assertDictEqual(salt.utils.args.clean_kwargs(__pub_foo="bar"), {})
        self.assertDictEqual(salt.utils.args.clean_kwargs(__foo_bar="gwar"), {})
        self.assertDictEqual(
            salt.utils.args.clean_kwargs(foo_bar="gwar"), {"foo_bar": "gwar"}
        )

    def test_get_function_argspec(self):
        class DummyClass:
            def __init__(self, first):
                pass

            def __call__(self, first):
                pass

        dummy_class = DummyClass("foo")

        def dummy_func_nodefault(first, second, third):
            pass

        def dummy_func_default(first, second, third="fourth"):
            pass

        def dummy_func_varargs_keywords(*args, **kwargs):
            pass

        _ArgSpec = namedtuple("ArgSpec", "args varargs keywords defaults")

        # Callable class instance
        expected_argspec = _ArgSpec(
            args=["first"],
            varargs=None,
            keywords=None,
            defaults=None,
        )
        ret = salt.utils.args.get_function_argspec(dummy_class)
        self.assertEqual(ret, expected_argspec)

        # Function with no varargs, no keywords, and no defaults
        expected_argspec = _ArgSpec(
            args=["first", "second", "third"],
            varargs=None,
            keywords=None,
            defaults=None,
        )
        ret = salt.utils.args.get_function_argspec(dummy_func_nodefault)
        self.assertEqual(ret, expected_argspec)

        # Function with no varargs, no keywords, and one default
        expected_argspec = _ArgSpec(
            args=["first", "second", "third"],
            varargs=None,
            keywords=None,
            defaults=("fourth",),
        )
        ret = salt.utils.args.get_function_argspec(dummy_func_default)
        self.assertEqual(ret, expected_argspec)

        # Function with both varargs and keywords
        expected_argspec = _ArgSpec(
            args=[],
            varargs="args",
            keywords="kwargs",
            defaults=None,
        )
        ret = salt.utils.args.get_function_argspec(dummy_func_varargs_keywords)
        self.assertEqual(ret, expected_argspec)

        # Test that is_class_method=True is respected. Note that we don't
        # actually have a decorated function from rest_tornado here to test
        # this case, but we're testing for the behavior we expect, which is
        # that the first argument is popped off of the args.
        expected_argspec = _ArgSpec(
            args=["second", "third"],
            varargs=None,
            keywords=None,
            defaults=None,
        )
        ret = salt.utils.args.get_function_argspec(
            dummy_func_nodefault, is_class_method=True
        )
        self.assertEqual(ret, expected_argspec)

    def test_parse_kwarg(self):
        ret = salt.utils.args.parse_kwarg("foo=bar")
        self.assertEqual(ret, ("foo", "bar"))

        ret = salt.utils.args.parse_kwarg("foobar")
        self.assertEqual(ret, (None, None))

    def test_arg_lookup(self):
        def dummy_func(first, second, third, fourth="fifth"):
            pass

        expected_dict = {
            "args": ["first", "second", "third"],
            "kwargs": {"fourth": "fifth"},
        }
        ret = salt.utils.args.arg_lookup(dummy_func)
        self.assertEqual(expected_dict, ret)

    def test_format_call(self):
        with patch("salt.utils.args.arg_lookup") as arg_lookup:

            def dummy_func(first=None, second=None, third=None):
                pass

            arg_lookup.return_value = {
                "args": ["first", "second", "third"],
                "kwargs": {},
            }
            get_function_argspec = DEFAULT
            get_function_argspec.return_value = namedtuple(
                "ArgSpec", "args varargs keywords defaults"
            )(
                args=["first", "second", "third", "fourth"],
                varargs=None,
                keywords=None,
                defaults=("fifth",),
            )

            # Make sure we raise an error if we don't pass in the requisite number of arguments
            self.assertRaises(
                SaltInvocationError, salt.utils.args.format_call, dummy_func, {"1": 2}
            )

            # Make sure we warn on invalid kwargs
            self.assertRaises(
                SaltInvocationError,
                salt.utils.args.format_call,
                dummy_func,
                {"first": 2, "seconds": 2, "third": 3},
            )

            ret = salt.utils.args.format_call(
                dummy_func,
                {"first": 2, "second": 2, "third": 3},
                expected_extra_kws=("first", "second", "third"),
            )
            self.assertDictEqual(ret, {"args": [], "kwargs": {}})

    def test_format_call_simple_args(self):
        def foo(one, two=2, three=3):
            pass

        self.assertEqual(
            salt.utils.args.format_call(foo, dict(one=10, two=20, three=30)),
            {"args": [10], "kwargs": dict(two=20, three=30)},
        )
        self.assertEqual(
            salt.utils.args.format_call(foo, dict(one=10, two=20)),
            {"args": [10], "kwargs": dict(two=20, three=3)},
        )
        self.assertEqual(
            salt.utils.args.format_call(foo, dict(one=2)),
            {"args": [2], "kwargs": dict(two=2, three=3)},
        )

    def test_format_call_mimic_typeerror_exceptions(self):
        def foo(one, two=2, three=3):
            pass

        def foo2(one, two, three=3):
            pass

        with self.assertRaisesRegex(
            SaltInvocationError, r"foo takes at least 1 argument \(0 given\)"
        ):
            salt.utils.args.format_call(foo, dict(two=3))

        with self.assertRaisesRegex(
            TypeError, r"foo2 takes at least 2 arguments \(1 given\)"
        ):
            salt.utils.args.format_call(foo2, dict(one=1))

    def test_argspec_report(self):
        def _test_spec(arg1, arg2, kwarg1=None):
            pass

        test_functions = {"test_module.test_spec": _test_spec}
        ret = salt.utils.args.argspec_report(test_functions, "test_module.test_spec")
        self.assertDictEqual(
            ret,
            {
                "test_module.test_spec": {
                    "kwargs": None,
                    "args": ["arg1", "arg2", "kwarg1"],
                    "defaults": (None,),
                    "varargs": None,
                }
            },
        )

    def test_test_mode(self):
        self.assertTrue(salt.utils.args.test_mode(test=True))
        self.assertTrue(salt.utils.args.test_mode(Test=True))
        self.assertTrue(salt.utils.args.test_mode(tEsT=True))

    def test_parse_function_no_args(self):
        fun, args, kwargs = salt.utils.args.parse_function("amod.afunc()")
        self.assertEqual(fun, "amod.afunc")
        self.assertEqual(args, [])
        self.assertEqual(kwargs, {})

    def test_parse_function_args_only(self):
        fun, args, kwargs = salt.utils.args.parse_function("amod.afunc(str1, str2)")
        self.assertEqual(fun, "amod.afunc")
        self.assertEqual(args, ["str1", "str2"])
        self.assertEqual(kwargs, {})

    def test_parse_function_kwargs_only(self):
        fun, args, kwargs = salt.utils.args.parse_function(
            "amod.afunc(kw1=val1, kw2=val2)"
        )
        self.assertEqual(fun, "amod.afunc")
        self.assertEqual(args, [])
        self.assertEqual(kwargs, {"kw1": "val1", "kw2": "val2"})

    def test_parse_function_args_kwargs(self):
        fun, args, kwargs = salt.utils.args.parse_function(
            "amod.afunc(str1, str2, kw1=val1, kw2=val2)"
        )
        self.assertEqual(fun, "amod.afunc")
        self.assertEqual(args, ["str1", "str2"])
        self.assertEqual(kwargs, {"kw1": "val1", "kw2": "val2"})

    def test_parse_function_malformed_no_name(self):
        fun, args, kwargs = salt.utils.args.parse_function(
            "(str1, str2, kw1=val1, kw2=val2)"
        )
        self.assertIsNone(fun)
        self.assertIsNone(args)
        self.assertIsNone(kwargs)

    def test_parse_function_malformed_not_fun_def(self):
        fun, args, kwargs = salt.utils.args.parse_function("foo bar, some=text")
        self.assertIsNone(fun)
        self.assertIsNone(args)
        self.assertIsNone(kwargs)

    def test_parse_function_wrong_bracket_style(self):
        fun, args, kwargs = salt.utils.args.parse_function(
            "amod.afunc[str1, str2, kw1=val1, kw2=val2]"
        )
        self.assertIsNone(fun)
        self.assertIsNone(args)
        self.assertIsNone(kwargs)

    def test_parse_function_brackets_unballanced(self):
        fun, args, kwargs = salt.utils.args.parse_function(
            "amod.afunc(str1, str2, kw1=val1, kw2=val2"
        )
        self.assertIsNone(fun)
        self.assertIsNone(args)
        self.assertIsNone(kwargs)
        fun, args, kwargs = salt.utils.args.parse_function(
            "amod.afunc(str1, str2, kw1=val1, kw2=val2]"
        )
        self.assertIsNone(fun)
        self.assertIsNone(args)
        self.assertIsNone(kwargs)
        fun, args, kwargs = salt.utils.args.parse_function(
            "amod.afunc(str1, str2, kw1=(val1[val2)], kw2=val2)"
        )
        self.assertIsNone(fun)
        self.assertIsNone(args)
        self.assertIsNone(kwargs)

    def test_parse_function_brackets_in_quotes(self):
        fun, args, kwargs = salt.utils.args.parse_function(
            'amod.afunc(str1, str2, kw1="(val1[val2)]", kw2=val2)'
        )
        self.assertEqual(fun, "amod.afunc")
        self.assertEqual(args, ["str1", "str2"])
        self.assertEqual(kwargs, {"kw1": "(val1[val2)]", "kw2": "val2"})

    def test_parse_function_quotes(self):
        fun, args, kwargs = salt.utils.args.parse_function(
            "amod.afunc(\"double \\\" single '\", 'double \" single \\'',"
            ' kw1="equal=equal", kw2=val2)'
        )
        self.assertEqual(fun, "amod.afunc")
        self.assertEqual(args, ["double \" single '", "double \" single '"])
        self.assertEqual(kwargs, {"kw1": "equal=equal", "kw2": "val2"})

    def test_yamlify_arg(self):
        """
        Test that we properly yamlify CLI input. In several of the tests below
        assertIs is used instead of assertEqual. This is because we want to
        confirm that the return value is not a copy of the original, but the
        same instance as the original.
        """

        def _yamlify_arg(item):
            log.debug("Testing yamlify_arg with %r", item)
            return salt.utils.args.yamlify_arg(item)

        # Make sure non-strings are just returned back
        for item in (True, False, None, 123, 45.67, ["foo"], {"foo": "bar"}):
            self.assertIs(_yamlify_arg(item), item)

        # Make sure whitespace-only isn't loaded as None
        for item in ("", "\t", " "):
            self.assertIs(_yamlify_arg(item), item)

        # This value would be loaded as an int (123), the underscores would be
        # ignored. Test that we identify this case and return the original
        # value.
        item = "1_2_3"
        self.assertIs(_yamlify_arg(item), item)

        # The '#' is treated as a comment when not part of a data structure, we
        # don't want that behavior
        for item in ("# hash at beginning", "Hello world! # hash elsewhere"):
            self.assertIs(_yamlify_arg(item), item)

        # However we _do_ want the # to be intact if it _is_ within a data
        # structure.
        item = '["foo", "bar", "###"]'
        self.assertEqual(_yamlify_arg(item), ["foo", "bar", "###"])
        item = '{"foo": "###"}'
        self.assertEqual(_yamlify_arg(item), {"foo": "###"})

        # The string "None" should load _as_ None
        self.assertIs(_yamlify_arg("None"), None)

        # Leading dashes, or strings containing colons, will result in lists
        # and dicts, and we only want to load lists and dicts when the strings
        # look like data structures.
        for item in ("- foo", "foo: bar"):
            self.assertIs(_yamlify_arg(item), item)

        # Make sure we don't load '|' as ''
        item = "|"
        self.assertIs(_yamlify_arg(item), item)

        # Make sure we don't load '!' as something else (None in 2018.3, '' in newer)
        item = "!"
        self.assertIs(_yamlify_arg(item), item)

        # Make sure we load ints, floats, and strings correctly
        self.assertEqual(_yamlify_arg("123"), 123)
        self.assertEqual(_yamlify_arg("45.67"), 45.67)
        self.assertEqual(_yamlify_arg("foo"), "foo")

        # We tested list/dict loading above, but there is separate logic when
        # the string contains a '#', so we need to test again here.
        self.assertEqual(_yamlify_arg('["foo", "bar"]'), ["foo", "bar"])
        self.assertEqual(_yamlify_arg('{"foo": "bar"}'), {"foo": "bar"})

        # Make sure that an empty string is loaded properly.
        self.assertEqual(_yamlify_arg("   "), "   ")

        # Make sure that we don't improperly load strings that would be
        # interpreted by PyYAML as YAML document start/end.
        self.assertEqual(_yamlify_arg("---"), "---")
        self.assertEqual(_yamlify_arg("--- "), "--- ")
        self.assertEqual(_yamlify_arg("..."), "...")
        self.assertEqual(_yamlify_arg(" ..."), " ...")

        # Make sure that non-printable whitespace is not YAML-loaded
        self.assertEqual(_yamlify_arg("foo\t\nbar"), "foo\t\nbar")


class KwargRegexTest(TestCase):
    def test_arguments_regex(self):
        argument_matches = (
            ("pip=1.1", ("pip", "1.1")),
            ("pip==1.1", None),
            ("pip=1.2=1", ("pip", "1.2=1")),
        )
        for argument, match in argument_matches:
            if match is None:
                self.assertIsNone(salt.utils.args.KWARG_REGEX.match(argument))
            else:
                self.assertEqual(
                    salt.utils.args.KWARG_REGEX.match(argument).groups(), match
                )



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest
import salt.states.pkgng as pkgng
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {pkgng: {}}


def test_update_packaging_site():
    """
    Test to execute update_packaging_site.
    """
    name = "http://192.168.0.2"

    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    mock_t = MagicMock(return_value=True)
    with patch.dict(pkgng.__salt__, {"pkgng.update_package_site": mock_t}):
        assert pkgng.update_packaging_site(name) == ret



"""
    :synopsis: Unit Tests for Windows PKI Module 'module.win_pki'
    :platform: Windows
    :maturity: develop
    .. versionadded:: 2017.7.0
"""


import salt.modules.win_pki as win_pki
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase

CERT_PATH = r"C:\certs\testdomain.local.cer"
THUMBPRINT = "9988776655443322111000AAABBBCCCDDDEEEFFF"

CERTS = {
    THUMBPRINT: {
        "dnsnames": ["testdomain.local"],
        "serialnumber": "0123456789AABBCCDD",
        "subject": "CN=testdomain.local, OU=testou, O=testorg, S=California, C=US",
        "thumbprint": THUMBPRINT,
        "version": 3,
    }
}

STORES = {
    "CurrentUser": [
        "AuthRoot",
        "CA",
        "ClientAuthIssuer",
        "Disallowed",
        "MSIEHistoryJournal",
        "My",
        "Root",
        "SmartCardRoot",
        "Trust",
        "TrustedPeople",
        "TrustedPublisher",
        "UserDS",
    ],
    "LocalMachine": [
        "AuthRoot",
        "CA",
        "ClientAuthIssuer",
        "Disallowed",
        "My",
        "Remote Desktop",
        "Root",
        "SmartCardRoot",
        "Trust",
        "TrustedDevices",
        "TrustedPeople",
        "TrustedPublisher",
        "WebHosting",
    ],
}

JSON_CERTS = [
    {
        "DnsNameList": [
            {"Punycode": "testdomain.local", "Unicode": "testdomain.local"}
        ],
        "SerialNumber": "0123456789AABBCCDD",
        "Subject": "CN=testdomain.local, OU=testou, O=testorg, S=California, C=US",
        "Thumbprint": "9988776655443322111000AAABBBCCCDDDEEEFFF",
        "Version": 3,
    }
]

JSON_STORES = [
    {"LocationName": "CurrentUser", "StoreNames": STORES["CurrentUser"]},
    {"LocationName": "LocalMachine", "StoreNames": STORES["LocalMachine"]},
]


class WinPkiTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.win_pki
    """

    def setup_loader_modules(self):
        return {win_pki: {}}

    def test_get_stores(self):
        """
        Test - Get the certificate location contexts and their corresponding stores.
        """
        with patch.dict(win_pki.__salt__), patch(
            "salt.modules.win_pki._cmd_run", MagicMock(return_value=JSON_STORES)
        ):
            self.assertEqual(win_pki.get_stores(), STORES)

    def test_get_certs(self):
        """
        Test - Get the available certificates in the given store.
        """
        with patch.dict(win_pki.__salt__), patch(
            "salt.modules.win_pki._cmd_run", MagicMock(return_value=JSON_CERTS)
        ), patch(
            "salt.modules.win_pki._validate_cert_path", MagicMock(return_value=None)
        ):
            self.assertEqual(win_pki.get_certs(), CERTS)

    def test_get_cert_file(self):
        """
        Test - Get the details of the certificate file.
        """
        kwargs = {"name": CERT_PATH}
        with patch.dict(win_pki.__salt__), patch(
            "os.path.isfile", MagicMock(return_value=True)
        ), patch("salt.modules.win_pki._cmd_run", MagicMock(return_value=JSON_CERTS)):
            self.assertEqual(win_pki.get_cert_file(**kwargs), CERTS[THUMBPRINT])

    def test_import_cert(self):
        """
        Test - Import the certificate file into the given certificate store.
        """
        kwargs = {"name": CERT_PATH}
        mock_value = MagicMock(return_value=CERT_PATH)
        with patch.dict(win_pki.__salt__, {"cp.cache_file": mock_value}), patch(
            "salt.modules.win_pki._cmd_run", MagicMock(return_value=JSON_CERTS)
        ), patch(
            "salt.modules.win_pki._validate_cert_path", MagicMock(return_value=None)
        ), patch(
            "salt.modules.win_pki.get_cert_file",
            MagicMock(return_value=CERTS[THUMBPRINT]),
        ), patch(
            "salt.modules.win_pki.get_certs", MagicMock(return_value=CERTS)
        ):
            self.assertTrue(win_pki.import_cert(**kwargs))

    def test_export_cert(self):
        """
        Test - Export the certificate to a file from the given certificate store.
        """
        kwargs = {"name": CERT_PATH, "thumbprint": THUMBPRINT}
        with patch.dict(win_pki.__salt__), patch(
            "salt.modules.win_pki._cmd_run", MagicMock(return_value="True")
        ), patch(
            "salt.modules.win_pki._validate_cert_format", MagicMock(return_value=None)
        ), patch(
            "salt.modules.win_pki._validate_cert_path", MagicMock(return_value=None)
        ):
            self.assertTrue(win_pki.export_cert(**kwargs))

    def test_test_cert(self):
        """
        Test - Check the certificate for validity.
        """
        with patch.dict(win_pki.__salt__), patch(
            "salt.modules.win_pki._cmd_run", MagicMock(return_value="True")
        ), patch(
            "salt.modules.win_pki._validate_cert_path", MagicMock(return_value=None)
        ):
            self.assertTrue(win_pki.test_cert(thumbprint=THUMBPRINT))

    def test_remove_cert(self):
        """
        Test - Remove the certificate from the given certificate store.
        """
        with patch.dict(win_pki.__salt__), patch(
            "salt.modules.win_pki._cmd_run", MagicMock(return_value=None)
        ), patch(
            "salt.modules.win_pki._validate_cert_path", MagicMock(return_value=None)
        ), patch(
            "salt.modules.win_pki.get_certs", MagicMock(return_value=CERTS)
        ):
            self.assertTrue(win_pki.remove_cert(thumbprint=THUMBPRINT[::-1]))



import pytest
import salt.config
import salt.version

pytestmark = [
    pytest.mark.slow_test,
]


def test_ping(
    mm_master_1_salt_cli, salt_mm_minion_1, mm_master_2_salt_cli, salt_mm_minion_2
):
    """
    test.ping
    """
    ret = mm_master_1_salt_cli.run("test.ping", minion_tgt=salt_mm_minion_1.id)
    assert ret.exitcode == 0
    assert ret.json is True

    ret = mm_master_2_salt_cli.run("test.ping", minion_tgt=salt_mm_minion_1.id)
    assert ret.exitcode == 0
    assert ret.json is True

    ret = mm_master_1_salt_cli.run("test.ping", minion_tgt=salt_mm_minion_2.id)
    assert ret.exitcode == 0
    assert ret.json is True

    ret = mm_master_2_salt_cli.run("test.ping", minion_tgt=salt_mm_minion_2.id)
    assert ret.exitcode == 0
    assert ret.json is True


def test_echo(
    mm_master_1_salt_cli, salt_mm_minion_1, mm_master_2_salt_cli, salt_mm_minion_2
):
    """
    test.echo
    """
    ret = mm_master_1_salt_cli.run("test.echo", "text", minion_tgt=salt_mm_minion_1.id)
    assert ret.exitcode == 0
    assert ret.json == "text"

    ret = mm_master_2_salt_cli.run("test.echo", "text", minion_tgt=salt_mm_minion_1.id)
    assert ret.exitcode == 0
    assert ret.json == "text"

    ret = mm_master_1_salt_cli.run("test.echo", "text", minion_tgt=salt_mm_minion_2.id)
    assert ret.exitcode == 0
    assert ret.json == "text"

    ret = mm_master_2_salt_cli.run("test.echo", "text", minion_tgt=salt_mm_minion_2.id)
    assert ret.exitcode == 0
    assert ret.json == "text"


def test_version(
    mm_master_1_salt_cli, salt_mm_minion_1, mm_master_2_salt_cli, salt_mm_minion_2
):
    """
    test.version
    """
    ret = mm_master_1_salt_cli.run("test.version", minion_tgt=salt_mm_minion_1.id)
    assert ret.exitcode == 0
    assert ret.json == salt.version.__saltstack_version__.string

    ret = mm_master_2_salt_cli.run("test.version", minion_tgt=salt_mm_minion_1.id)
    assert ret.exitcode == 0
    assert ret.json == salt.version.__saltstack_version__.string

    ret = mm_master_1_salt_cli.run("test.version", minion_tgt=salt_mm_minion_2.id)
    assert ret.exitcode == 0
    assert ret.json == salt.version.__saltstack_version__.string

    ret = mm_master_2_salt_cli.run("test.version", minion_tgt=salt_mm_minion_2.id)
    assert ret.exitcode == 0
    assert ret.json == salt.version.__saltstack_version__.string


def test_conf_test(
    mm_master_1_salt_cli, salt_mm_minion_1, mm_master_2_salt_cli, salt_mm_minion_2
):
    """
    test.conf_text
    """
    ret = mm_master_1_salt_cli.run("test.conf_test", minion_tgt=salt_mm_minion_1.id)
    assert ret.exitcode == 0
    assert ret.json == "baz"

    ret = mm_master_2_salt_cli.run("test.conf_test", minion_tgt=salt_mm_minion_1.id)
    assert ret.exitcode == 0
    assert ret.json == "baz"

    ret = mm_master_1_salt_cli.run("test.conf_test", minion_tgt=salt_mm_minion_2.id)
    assert ret.exitcode == 0
    assert ret.json == "baz"

    ret = mm_master_2_salt_cli.run("test.conf_test", minion_tgt=salt_mm_minion_2.id)
    assert ret.exitcode == 0
    assert ret.json == "baz"


def test_cross_test(
    mm_master_1_salt_cli, salt_mm_minion_1, mm_master_2_salt_cli, salt_mm_minion_2
):
    """
    test.cross_text
    """
    ret = mm_master_1_salt_cli.run(
        "test.cross_test", "test.ping", minion_tgt=salt_mm_minion_1.id
    )
    assert ret.exitcode == 0
    assert ret.json is True

    ret = mm_master_2_salt_cli.run(
        "test.cross_test", "test.ping", minion_tgt=salt_mm_minion_1.id
    )
    assert ret.exitcode == 0
    assert ret.json is True

    ret = mm_master_1_salt_cli.run(
        "test.cross_test", "test.ping", minion_tgt=salt_mm_minion_2.id
    )
    assert ret.exitcode == 0
    assert ret.json is True

    ret = mm_master_2_salt_cli.run(
        "test.cross_test", "test.ping", minion_tgt=salt_mm_minion_2.id
    )
    assert ret.exitcode == 0
    assert ret.json is True


def test_outputter(
    mm_master_1_salt_cli, salt_mm_minion_1, mm_master_2_salt_cli, salt_mm_minion_2
):
    """
    test.outputter
    """
    ret = mm_master_1_salt_cli.run(
        "test.outputter", "text", minion_tgt=salt_mm_minion_1.id
    )
    assert ret.exitcode == 0
    assert ret.json == "text"

    ret = mm_master_2_salt_cli.run(
        "test.outputter", "text", minion_tgt=salt_mm_minion_1.id
    )
    assert ret.exitcode == 0
    assert ret.json == "text"

    ret = mm_master_1_salt_cli.run(
        "test.outputter", "text", minion_tgt=salt_mm_minion_2.id
    )
    assert ret.exitcode == 0
    assert ret.json == "text"

    ret = mm_master_2_salt_cli.run(
        "test.outputter", "text", minion_tgt=salt_mm_minion_2.id
    )
    assert ret.exitcode == 0
    assert ret.json == "text"


def test_fib(
    mm_master_1_salt_cli, salt_mm_minion_1, mm_master_2_salt_cli, salt_mm_minion_2
):
    """
    test.fib
    """
    ret = mm_master_1_salt_cli.run("test.fib", "20", minion_tgt=salt_mm_minion_1.id)
    assert ret.exitcode == 0
    assert ret.json[0] == 6765

    ret = mm_master_2_salt_cli.run("test.fib", "20", minion_tgt=salt_mm_minion_1.id)
    assert ret.exitcode == 0
    assert ret.json[0] == 6765

    ret = mm_master_1_salt_cli.run("test.fib", "20", minion_tgt=salt_mm_minion_2.id)
    assert ret.exitcode == 0
    assert ret.json[0] == 6765

    ret = mm_master_2_salt_cli.run("test.fib", "20", minion_tgt=salt_mm_minion_2.id)
    assert ret.exitcode == 0
    assert ret.json[0] == 6765


def test_collatz(
    mm_master_1_salt_cli, salt_mm_minion_1, mm_master_2_salt_cli, salt_mm_minion_2
):
    """
    test.fib
    """
    ret = mm_master_1_salt_cli.run("test.collatz", "40", minion_tgt=salt_mm_minion_1.id)
    assert ret.exitcode == 0
    assert ret.json[0][-1] == 2

    ret = mm_master_2_salt_cli.run("test.collatz", "40", minion_tgt=salt_mm_minion_1.id)
    assert ret.exitcode == 0
    assert ret.json[0][-1] == 2

    ret = mm_master_1_salt_cli.run("test.collatz", "40", minion_tgt=salt_mm_minion_2.id)
    assert ret.exitcode == 0
    assert ret.json[0][-1] == 2

    ret = mm_master_2_salt_cli.run("test.collatz", "40", minion_tgt=salt_mm_minion_2.id)
    assert ret.exitcode == 0
    assert ret.json[0][-1] == 2


def test_get_opts(
    mm_master_1_salt_cli, salt_mm_minion_1, mm_master_2_salt_cli, salt_mm_minion_2
):
    """
    test.conf_text
    """
    ret = mm_master_1_salt_cli.run("test.get_opts", minion_tgt=salt_mm_minion_1.id)
    assert ret.exitcode == 0
    assert ret.json["cachedir"] == salt_mm_minion_1.config["cachedir"]

    ret = mm_master_2_salt_cli.run("test.get_opts", minion_tgt=salt_mm_minion_1.id)
    assert ret.exitcode == 0
    assert ret.json["cachedir"] == salt_mm_minion_1.config["cachedir"]

    ret = mm_master_1_salt_cli.run("test.get_opts", minion_tgt=salt_mm_minion_2.id)
    assert ret.exitcode == 0
    assert ret.json["cachedir"] == salt_mm_minion_2.config["cachedir"]

    ret = mm_master_2_salt_cli.run("test.get_opts", minion_tgt=salt_mm_minion_2.id)
    assert ret.exitcode == 0
    assert ret.json["cachedir"] == salt_mm_minion_2.config["cachedir"]



"""
    :codeauthor: Rahul Handay <rahulha@saltstack.com>
"""

import logging

import pytest
import salt.states.user as user
from tests.support.mock import MagicMock, Mock, patch

log = logging.getLogger(__name__)


@pytest.fixture
def configure_loader_modules():
    return {user: {}}


def test_present():
    """
    Test to ensure that the named user is present with
    the specified properties
    """
    ret = {"name": "salt", "changes": {}, "result": False, "comment": ""}
    mock_false = MagicMock(return_value=False)
    mock_empty_list = MagicMock(return_value=[])
    with patch.dict(user.__grains__, {"kernel": "Linux"}):
        with patch.dict(
            user.__salt__,
            {
                "group.info": mock_false,
                "user.info": mock_empty_list,
                "user.chkey": mock_empty_list,
                "user.add": mock_false,
            },
        ):
            ret.update({"comment": "The following group(s) are not present: salt"})
            assert user.present("salt", groups=["salt"]) == ret

            mock_false = MagicMock(
                side_effect=[
                    {"key": "value"},
                    {"key": "value"},
                    {"key": "value"},
                    False,
                    False,
                ]
            )
            with patch.object(user, "_changes", mock_false):
                with patch.dict(user.__opts__, {"test": True}):
                    ret.update(
                        {
                            "comment": (
                                "The following user attributes are set to be changed:\n"
                                "key: value\n"
                            ),
                            "result": None,
                        }
                    )
                    assert user.present("salt") == ret

                with patch.dict(user.__opts__, {"test": False}):
                    comment = "These values could not be changed: {!r}".format(
                        {"key": "value"}
                    )
                    ret.update({"comment": comment, "result": False})
                    assert user.present("salt") == ret

                    with patch.dict(user.__opts__, {"test": True}):
                        ret.update(
                            {"comment": "User salt set to be added", "result": None}
                        )
                        assert user.present("salt") == ret

                    with patch.dict(user.__opts__, {"test": False}):
                        ret.update(
                            {
                                "comment": "Failed to create new user salt",
                                "result": False,
                            }
                        )
                        assert user.present("salt") == ret


def test_present_invalid_uid_change():
    mock_info = MagicMock(
        side_effect=[
            {
                "uid": 5000,
                "gid": 5000,
                "groups": ["foo"],
                "home": "/home/foo",
                "fullname": "Foo Bar",
            }
        ]
    )
    dunder_salt = {
        "user.info": mock_info,
        "file.group_to_gid": MagicMock(side_effect=["foo"]),
        "file.gid_to_group": MagicMock(side_effect=[5000, 5000]),
    }
    with patch.dict(user.__grains__, {"kernel": "Linux"}), patch.dict(
        user.__salt__, dunder_salt
    ):
        ret = user.present("foo", uid=5001)
        assert not ret["result"]
        assert ret["comment"].count("not permitted") == 1


def test_present_invalid_gid_change():
    mock_info = MagicMock(
        side_effect=[
            {
                "uid": 5000,
                "gid": 5000,
                "groups": ["foo"],
                "home": "/home/foo",
                "fullname": "Foo Bar",
            }
        ]
    )
    dunder_salt = {
        "user.info": mock_info,
        "file.group_to_gid": MagicMock(side_effect=["foo"]),
        "file.gid_to_group": MagicMock(side_effect=[5000, 5000]),
    }
    with patch.dict(user.__grains__, {"kernel": "Linux"}), patch.dict(
        user.__salt__, dunder_salt
    ):
        ret = user.present("foo", gid=5001)
        assert not ret["result"]
        assert ret["comment"].count("not permitted") == 1


def test_present_invalid_uid_gid_change():
    mock_info = MagicMock(
        side_effect=[
            {
                "uid": 5000,
                "gid": 5000,
                "groups": ["foo"],
                "home": "/home/foo",
                "fullname": "Foo Bar",
            }
        ]
    )
    dunder_salt = {
        "user.info": mock_info,
        "file.group_to_gid": MagicMock(side_effect=["foo"]),
        "file.gid_to_group": MagicMock(side_effect=[5000, 5000]),
    }
    with patch.dict(user.__grains__, {"kernel": "Linux"}), patch.dict(
        user.__salt__, dunder_salt
    ):
        ret = user.present("foo", uid=5001, gid=5001)
        assert not ret["result"]
        assert ret["comment"].count("not permitted") == 2


def test_present_uid_gid_change():
    before = {
        "uid": 5000,
        "gid": 5000,
        "groups": ["foo"],
        "home": "/home/foo",
        "fullname": "Foo Bar",
    }
    after = {
        "uid": 5001,
        "gid": 5001,
        "groups": ["othergroup"],
        "home": "/home/foo",
        "fullname": "Foo Bar",
    }
    # user.info should be called 4 times. Once the first time that
    # _changes() is called, once before and after changes are applied (to
    # get the before/after for the changes dict, and one last time to
    # confirm that no changes still need to be made.
    mock_info = MagicMock(side_effect=[before, before, after, after])
    mock_group_to_gid = MagicMock(side_effect=["foo", "othergroup"])
    mock_gid_to_group = MagicMock(side_effect=[5000, 5000, 5001, 5001])
    dunder_salt = {
        "user.info": mock_info,
        "user.chuid": Mock(),
        "user.chgid": Mock(),
        "file.group_to_gid": mock_group_to_gid,
        "file.gid_to_group": mock_gid_to_group,
    }
    with patch.dict(user.__grains__, {"kernel": "Linux"}), patch.dict(
        user.__salt__, dunder_salt
    ), patch.dict(user.__opts__, {"test": False}), patch(
        "os.path.isdir", MagicMock(return_value=True)
    ):
        ret = user.present(
            "foo", uid=5001, gid=5001, allow_uid_change=True, allow_gid_change=True
        )
        assert ret == {
            "comment": "Updated user foo",
            "changes": {"gid": 5001, "uid": 5001, "groups": ["othergroup"]},
            "name": "foo",
            "result": True,
        }


def test_absent():
    """
    Test to ensure that the named user is absent
    """
    ret = {"name": "salt", "changes": {}, "result": None, "comment": ""}
    mock = MagicMock(side_effect=[True, True, False])
    mock1 = MagicMock(return_value=False)
    with patch.dict(
        user.__salt__,
        {"user.info": mock, "user.delete": mock1, "group.info": mock1},
    ):
        with patch.dict(user.__opts__, {"test": True}):
            ret.update({"comment": "User salt set for removal"})
            assert user.absent("salt") == ret

        with patch.dict(user.__opts__, {"test": False}):
            ret.update({"comment": "Failed to remove user salt", "result": False})
            assert user.absent("salt") == ret

        ret.update({"comment": "User salt is not present", "result": True})
        assert user.absent("salt") == ret


def test_changes():
    """
    Test salt.states.user._changes
    """
    mock_info = MagicMock(
        return_value={
            "uid": 5000,
            "gid": 5000,
            "groups": ["foo"],
            "home": "/home/foo",
            "fullname": "Foo Bar",
        }
    )
    shadow_info = MagicMock(
        return_value={"min": 2, "max": 88888, "inact": 77, "warn": 14}
    )
    shadow_hash = MagicMock(return_value="abcd")
    dunder_salt = {
        "user.info": mock_info,
        "shadow.info": shadow_info,
        "shadow.default_hash": shadow_hash,
        "file.group_to_gid": MagicMock(side_effect=["foo"]),
        "file.gid_to_group": MagicMock(side_effect=[5000, 5000]),
    }

    def mock_exists(*args):
        return True

    with patch.dict(user.__grains__, {"kernel": "Linux"}), patch.dict(
        user.__salt__, dunder_salt
    ), patch.dict(user.__opts__, {"test": False}), patch("os.path.isdir", mock_exists):
        ret = user._changes("foo", maxdays=999999, inactdays=0, warndays=7)
        assert ret == {
            "maxdays": 999999,
            "mindays": 0,
            "fullname": "",
            "warndays": 7,
            "inactdays": 0,
        }



"""
Module for notifications via Twilio

.. versionadded:: 2014.7.0

:depends:   - twilio python module
:configuration: Configure this module by specifying the name of a configuration
    profile in the minion config, minion pillar, or master config (with :conf_master:`pillar_opts` set to True).

    .. warning: Setting pillar_opts to True in the master config may be considered
      unsafe as it copies the master config to pillar

    For example:

    .. code-block:: yaml

        my-twilio-account:
            twilio.account_sid: AC32a3c83990934481addd5ce1659f04d2
            twilio.auth_token: mytoken
"""

import logging

HAS_LIBS = False
try:
    import twilio

    # Grab version, ensure elements are ints
    twilio_version = tuple([int(x) for x in twilio.__version_info__])
    if twilio_version > (5,):
        TWILIO_5 = False
        from twilio.rest import Client as TwilioRestClient
        from twilio.rest import TwilioException as TwilioRestException
    else:
        TWILIO_5 = True
        from twilio.rest import TwilioRestClient
        from twilio import TwilioRestException  # pylint: disable=no-name-in-module
    HAS_LIBS = True
except ImportError:
    pass


log = logging.getLogger(__name__)

__virtualname__ = "twilio"


def __virtual__():
    """
    Only load this module if twilio is installed on this minion.
    """
    if HAS_LIBS:
        return __virtualname__
    return (
        False,
        "The twilio_notify execution module failed to load: the twilio python library"
        " is not installed.",
    )


def _get_twilio(profile):
    """
    Return the twilio connection
    """
    creds = __salt__["config.option"](profile)
    client = TwilioRestClient(
        creds.get("twilio.account_sid"),
        creds.get("twilio.auth_token"),
    )

    return client


def send_sms(profile, body, to, from_):
    """
    Send an sms

    CLI Example:

        twilio.send_sms my-twilio-account 'Test sms' '+18019999999' '+18011111111'
    """
    ret = {}
    ret["message"] = {}
    ret["message"]["sid"] = None
    client = _get_twilio(profile)
    try:
        if TWILIO_5:
            message = client.sms.messages.create(body=body, to=to, from_=from_)
        else:
            message = client.messages.create(body=body, to=to, from_=from_)
    except TwilioRestException as exc:
        ret["_error"] = {}
        ret["_error"]["code"] = exc.code
        ret["_error"]["msg"] = exc.msg
        ret["_error"]["status"] = exc.status
        log.debug("Could not send sms. Error: %s", ret)
        return ret
    ret["message"] = {}
    ret["message"]["sid"] = message.sid
    ret["message"]["price"] = message.price
    ret["message"]["price_unit"] = message.price_unit
    ret["message"]["status"] = message.status
    ret["message"]["num_segments"] = message.num_segments
    ret["message"]["body"] = message.body
    ret["message"]["date_sent"] = str(message.date_sent)
    ret["message"]["date_created"] = str(message.date_created)
    log.info(ret)
    return ret



"""
Management of OpenStack Keystone Roles
======================================

.. versionadded:: 2018.3.0

:depends: shade
:configuration: see :py:mod:`salt.modules.keystoneng` for setup instructions

Example States

.. code-block:: yaml

    create role:
      keystone_role.present:
        - name: role1

    delete role:
      keystone_role.absent:
        - name: role1

    create role with optional params:
      keystone_role.present:
        - name: role1
        - description: 'my group'
"""


__virtualname__ = "keystone_role"


def __virtual__():
    if "keystoneng.role_get" in __salt__:
        return __virtualname__
    return (
        False,
        "The keystoneng execution module failed to load: shade python module is not"
        " available",
    )


def present(name, auth=None, **kwargs):
    """
    Ensure an role exists

    name
        Name of the role

    description
        An arbitrary description of the role
    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    kwargs = __utils__["args.clean_kwargs"](**kwargs)

    __salt__["keystoneng.setup_clouds"](auth)

    kwargs["name"] = name
    role = __salt__["keystoneng.role_get"](**kwargs)

    if not role:
        if __opts__["test"] is True:
            ret["result"] = None
            ret["changes"] = kwargs
            ret["comment"] = "Role will be created."
            return ret

        role = __salt__["keystoneng.role_create"](**kwargs)
        ret["changes"]["id"] = role.id
        ret["changes"]["name"] = role.name
        ret["comment"] = "Created role"
        return ret
    # NOTE(SamYaple): Update support pending https://review.openstack.org/#/c/496992/
    return ret


def absent(name, auth=None, **kwargs):
    """
    Ensure role does not exist

    name
        Name of the role
    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    __salt__["keystoneng.setup_clouds"](auth)

    kwargs["name"] = name
    role = __salt__["keystoneng.role_get"](**kwargs)

    if role:
        if __opts__["test"] is True:
            ret["result"] = None
            ret["changes"] = {"id": role.id}
            ret["comment"] = "Role will be deleted."
            return ret

        __salt__["keystoneng.role_delete"](name=role)
        ret["changes"]["id"] = role.id
        ret["comment"] = "Deleted role"

    return ret



"""
Test case for the etcd SDB module
"""


import logging

import salt.sdb.etcd_db as etcd_db
import salt.utils.etcd_util as etcd_util
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, call, create_autospec, patch
from tests.support.unit import TestCase

log = logging.getLogger(__name__)


class TestEtcdSDB(LoaderModuleMockMixin, TestCase):
    """
    Test case for the etcd_db SDB module
    """

    def setup_loader_modules(self):
        return {
            etcd_db: {
                "__opts__": {
                    "myetcd": {
                        "url": "http://127.0.0.1",
                        "auth": {"token": "test", "method": "token"},
                    }
                }
            }
        }

    def setUp(self):
        self.instance = create_autospec(etcd_util.EtcdClient)
        self.EtcdClientMock = MagicMock()
        self.EtcdClientMock.return_value = self.instance

    def tearDown(self):
        del self.instance
        del self.EtcdClientMock

    def test_set(self):
        """
        Test salt.sdb.etcd_db.set function
        """
        with patch("salt.sdb.etcd_db._get_conn", self.EtcdClientMock):
            etcd_db.set_("sdb://myetcd/path/to/foo/bar", "super awesome")

        self.assertEqual(
            self.instance.set.call_args_list,
            [call("sdb://myetcd/path/to/foo/bar", "super awesome")],
        )

        self.assertEqual(
            self.instance.get.call_args_list,
            [call("sdb://myetcd/path/to/foo/bar")],
        )

    def test_get(self):
        """
        Test salt.sdb.etcd_db.get function
        """
        with patch("salt.sdb.etcd_db._get_conn", self.EtcdClientMock):
            etcd_db.get("sdb://myetcd/path/to/foo/bar")

        self.assertEqual(
            self.instance.get.call_args_list,
            [call("sdb://myetcd/path/to/foo/bar")],
        )



"""
Module for managing BCache sets

BCache is a block-level caching mechanism similar to ZFS L2ARC/ZIL, dm-cache and fscache.
It works by formatting one block device as a cache set, then adding backend devices
(which need to be formatted as such) to the set and activating them.

It's available in Linux mainline kernel since 3.10

https://www.kernel.org/doc/Documentation/bcache.txt

This module needs the bcache userspace tools to function.

.. versionadded:: 2016.3.0

"""

import logging
import os
import re
import time

import salt.utils.path

log = logging.getLogger(__name__)

LOG = {
    "trace": logging.TRACE,
    "debug": logging.DEBUG,
    "info": logging.INFO,
    "warn": logging.WARNING,
    "error": logging.ERROR,
    "crit": logging.CRITICAL,
}

__func_alias__ = {
    "attach_": "attach",
    "config_": "config",
    "super_": "super",
}

HAS_BLKDISCARD = salt.utils.path.which("blkdiscard") is not None


def __virtual__():
    """
    Only work when make-bcache is installed
    """
    return salt.utils.path.which("make-bcache") is not None


def uuid(dev=None):
    """
    Return the bcache UUID of a block device.
    If no device is given, the Cache UUID is returned.

    CLI Example:

    .. code-block:: bash

        salt '*' bcache.uuid
        salt '*' bcache.uuid /dev/sda
        salt '*' bcache.uuid bcache0

    """
    try:
        if dev is None:
            # take the only directory in /sys/fs/bcache and return its basename
            return list(salt.utils.path.os_walk("/sys/fs/bcache/"))[0][1][0]
        else:
            # basename of the /sys/block/{dev}/bcache/cache symlink target
            return os.path.basename(_bcsys(dev, "cache"))
    except Exception:  # pylint: disable=broad-except
        return False


def attach_(dev=None):
    """
    Attach a backing devices to a cache set
    If no dev is given, all backing devices will be attached.

    CLI Example:

    .. code-block:: bash

        salt '*' bcache.attach sdc
        salt '*' bcache.attach /dev/bcache1


    :return: bool or None if nuttin' happened
    """
    cache = uuid()
    if not cache:
        log.error("No cache to attach %s to", dev)
        return False

    if dev is None:
        res = {}
        for dev, data in status(alldevs=True).items():
            if "cache" in data:
                res[dev] = attach_(dev)

        return res if res else None

    bcache = uuid(dev)
    if bcache:
        if bcache == cache:
            log.info("%s is already attached to bcache %s, doing nothing", dev, cache)
            return None
        elif not detach(dev):
            return False

    log.debug("Attaching %s to bcache %s", dev, cache)

    if not _bcsys(
        dev,
        "attach",
        cache,
        "error",
        "Error attaching {} to bcache {}".format(dev, cache),
    ):
        return False

    return _wait(
        lambda: uuid(dev) == cache,
        "error",
        "{} received attach to bcache {}, but did not comply".format(dev, cache),
    )


def detach(dev=None):
    """
    Detach a backing device(s) from a cache set
    If no dev is given, all backing devices will be attached.

    Detaching a backing device will flush its write cache.
    This should leave the underlying device in a consistent state, but might take a while.

    CLI Example:

    .. code-block:: bash

        salt '*' bcache.detach sdc
        salt '*' bcache.detach bcache1

    """
    if dev is None:
        res = {}
        for dev, data in status(alldevs=True).items():
            if "cache" in data:
                res[dev] = detach(dev)

        return res if res else None

    log.debug("Detaching %s", dev)
    if not _bcsys(dev, "detach", "goaway", "error", "Error detaching {}".format(dev)):
        return False
    return _wait(
        lambda: uuid(dev) is False,
        "error",
        "{} received detach, but did not comply".format(dev),
        300,
    )


def start():
    """
    Trigger a start of the full bcache system through udev.

    CLI Example:

    .. code-block:: bash

        salt '*' bcache.start

    """
    if not _run_all("udevadm trigger", "error", "Error starting bcache: %s"):
        return False
    elif not _wait(
        lambda: uuid() is not False,
        "warn",
        "Bcache system started, but no active cache set found.",
    ):
        return False
    return True


def stop(dev=None):
    """
    Stop a bcache device
    If no device is given, all backing devices will be detached from the cache, which will subsequently be stopped.

    .. warning::
        'Stop' on an individual backing device means hard-stop;
        no attempt at flushing will be done and the bcache device will seemingly 'disappear' from the device lists

    CLI Example:

    .. code-block:: bash

        salt '*' bcache.stop

    """
    if dev is not None:
        log.warning("Stopping %s, device will only reappear after reregistering!", dev)
        if not _bcsys(dev, "stop", "goaway", "error", "Error stopping {}".format(dev)):
            return False
        return _wait(
            lambda: _sysfs_attr(_bcpath(dev)) is False,
            "error",
            "Device {} did not stop".format(dev),
            300,
        )
    else:
        cache = uuid()
        if not cache:
            log.warning("bcache already stopped?")
            return None

        if not _alltrue(detach()):
            return False
        elif not _fssys("stop", "goaway", "error", "Error stopping cache"):
            return False

        return _wait(lambda: uuid() is False, "error", "Cache did not stop", 300)


def back_make(dev, cache_mode="writeback", force=False, attach=True, bucket_size=None):
    """
    Create a backing device for attachment to a set.
    Because the block size must be the same, a cache set already needs to exist.

    CLI Example:

    .. code-block:: bash

        salt '*' bcache.back_make sdc cache_mode=writeback attach=True


    :param cache_mode: writethrough, writeback, writearound or none.
    :param force: Overwrite existing bcaches
    :param attach: Immediately attach the backing device to the set
    :param bucket_size: Size of a bucket (see kernel doc)
    """
    # pylint: disable=too-many-return-statements
    cache = uuid()

    if not cache:
        log.error("No bcache set found")
        return False
    elif _sysfs_attr(_bcpath(dev)):
        if not force:
            log.error(
                "%s already contains a bcache. Wipe it manually or use force", dev
            )
            return False
        elif uuid(dev) and not detach(dev):
            return False
        elif not stop(dev):
            return False

    dev = _devpath(dev)
    block_size = _size_map(_fssys("block_size"))
    # You might want to override, we pick the cache set's as sane default
    if bucket_size is None:
        bucket_size = _size_map(_fssys("bucket_size"))

    cmd = "make-bcache --block {} --bucket {} --{} --bdev {}".format(
        block_size, bucket_size, cache_mode, dev
    )
    if force:
        cmd += " --wipe-bcache"

    if not _run_all(cmd, "error", "Error creating backing device {}: %s".format(dev)):
        return False
    elif not _sysfs_attr(
        "fs/bcache/register",
        _devpath(dev),
        "error",
        "Error registering backing device {}".format(dev),
    ):
        return False
    elif not _wait(
        lambda: _sysfs_attr(_bcpath(dev)) is not False,
        "error",
        "Backing device {} did not register".format(dev),
    ):
        return False
    elif attach:
        return attach_(dev)

    return True


def cache_make(
    dev, reserved=None, force=False, block_size=None, bucket_size=None, attach=True
):
    """
    Create BCache cache on a block device.
    If blkdiscard is available the entire device will be properly cleared in advance.

    CLI Example:

    .. code-block:: bash

        salt '*' bcache.cache_make sdb reserved=10% block_size=4096


    :param reserved: if dev is a full device, create a partition table with this size empty.

        .. note::
              this increases the amount of reserved space available to SSD garbage collectors,
              potentially (vastly) increasing performance
    :param block_size: Block size of the cache; defaults to devices' logical block size
    :param force: Overwrite existing BCache sets
    :param attach: Attach all existing backend devices immediately
    """
    # TODO: multiple devs == md jbod

    # pylint: disable=too-many-return-statements
    # ---------------- Preflight checks ----------------
    cache = uuid()
    if cache:
        if not force:
            log.error("BCache cache %s is already on the system", cache)
            return False
        cache = _bdev()

    dev = _devbase(dev)
    udev = __salt__["udev.env"](dev)

    if (
        "ID_FS_TYPE" in udev
        or (udev.get("DEVTYPE", None) != "partition" and "ID_PART_TABLE_TYPE" in udev)
    ) and not force:
        log.error("%s already contains data, wipe first or force", dev)
        return False
    elif reserved is not None and udev.get("DEVTYPE", None) != "disk":
        log.error("Need a partitionable blockdev for reserved to work")
        return False

    _, block, bucket = _sizes(dev)

    if bucket_size is None:
        bucket_size = bucket
        # TODO: bucket from _sizes() makes no sense
        bucket_size = False
    if block_size is None:
        block_size = block

    # ---------------- Still here, start doing destructive stuff ----------------
    if cache:
        if not stop():
            return False
        # Wipe the current cache device as well,
        # forever ruining any chance of it accidentally popping up again
        elif not _wipe(cache):
            return False

    # Can't do enough wiping
    if not _wipe(dev):
        return False

    if reserved:
        cmd = (
            "parted -m -s -a optimal -- "
            "/dev/{0} mklabel gpt mkpart bcache-reserved 1M {1} mkpart bcache {1} 100%".format(
                dev, reserved
            )
        )
        # if wipe was incomplete & part layout remains the same,
        # this is one condition set where udev would make it accidentally popup again
        if not _run_all(
            cmd, "error", "Error creating bcache partitions on {}: %s".format(dev)
        ):
            return False
        dev = "{}2".format(dev)

    # ---------------- Finally, create a cache ----------------
    cmd = "make-bcache --cache /dev/{} --block {} --wipe-bcache".format(dev, block_size)

    # Actually bucket_size should always have a value, but for testing 0 is possible as well
    if bucket_size:
        cmd += " --bucket {}".format(bucket_size)

    if not _run_all(cmd, "error", "Error creating cache {}: %s".format(dev)):
        return False
    elif not _wait(
        lambda: uuid() is not False,
        "error",
        "Cache {} seemingly created OK, but FS did not activate".format(dev),
    ):
        return False

    if attach:
        return _alltrue(attach_())
    else:
        return True


def config_(dev=None, **kwargs):
    """
    Show or update config of a bcache device.

    If no device is given, operate on the cache set itself.

    CLI Example:

    .. code-block:: bash

        salt '*' bcache.config
        salt '*' bcache.config bcache1
        salt '*' bcache.config errors=panic journal_delay_ms=150
        salt '*' bcache.config bcache1 cache_mode=writeback writeback_percent=15

    :return: config or True/False
    """
    if dev is None:
        spath = _fspath()
    else:
        spath = _bcpath(dev)

    # filter out 'hidden' kwargs added by our favourite orchestration system
    updates = {key: val for key, val in kwargs.items() if not key.startswith("__")}

    if updates:
        endres = 0
        for key, val in updates.items():
            endres += _sysfs_attr(
                [spath, key],
                val,
                "warn",
                "Failed to update {} with {}".format(os.path.join(spath, key), val),
            )
        return endres > 0
    else:
        result = {}
        data = _sysfs_parse(spath, config=True, internals=True, options=True)
        for key in ("other_ro", "inter_ro"):
            if key in data:
                del data[key]

        for key in data:
            result.update(data[key])

        return result


def status(stats=False, config=False, internals=False, superblock=False, alldevs=False):
    """
    Show the full status of the BCache system and optionally all its involved devices

    CLI Example:

    .. code-block:: bash

        salt '*' bcache.status
        salt '*' bcache.status stats=True
        salt '*' bcache.status internals=True alldevs=True

    :param stats: include statistics
    :param config: include settings
    :param internals: include internals
    :param superblock: include superblock
    """
    bdevs = []
    for _, links, _ in salt.utils.path.os_walk("/sys/block/"):
        for block in links:
            if "bcache" in block:
                continue

            for spath, sdirs, _ in salt.utils.path.os_walk(
                "/sys/block/{}".format(block), followlinks=False
            ):
                if "bcache" in sdirs:
                    bdevs.append(os.path.basename(spath))
    statii = {}
    for bcache in bdevs:
        statii[bcache] = device(bcache, stats, config, internals, superblock)

    cuuid = uuid()
    cdev = _bdev()
    if cdev:
        count = 0
        for dev in statii:
            if dev != cdev:
                # it's a backing dev
                if statii[dev]["cache"] == cuuid:
                    count += 1
        statii[cdev]["attached_backing_devices"] = count

        if not alldevs:
            statii = statii[cdev]

    return statii


def device(dev, stats=False, config=False, internals=False, superblock=False):
    """
    Check the state of a single bcache device

    CLI Example:

    .. code-block:: bash

        salt '*' bcache.device bcache0
        salt '*' bcache.device /dev/sdc stats=True

    :param stats: include statistics
    :param settings: include all settings
    :param internals: include all internals
    :param superblock: include superblock info
    """
    result = {}

    if not _sysfs_attr(
        _bcpath(dev), None, "error", "{} is not a bcache fo any kind".format(dev)
    ):
        return False
    elif _bcsys(dev, "set"):
        # ---------------- It's the cache itself ----------------
        result["uuid"] = uuid()
        base_attr = [
            "block_size",
            "bucket_size",
            "cache_available_percent",
            "cache_replacement_policy",
            "congested",
        ]

        # ---------------- Parse through both the blockdev & the FS ----------------
        result.update(_sysfs_parse(_bcpath(dev), base_attr, stats, config, internals))
        result.update(_sysfs_parse(_fspath(), base_attr, stats, config, internals))

        result.update(result.pop("base"))
    else:
        # ---------------- It's a backing device ----------------
        back_uuid = uuid(dev)
        if back_uuid is not None:
            result["cache"] = back_uuid

        try:
            result["dev"] = os.path.basename(_bcsys(dev, "dev"))
        except Exception:  # pylint: disable=broad-except
            pass
        result["bdev"] = _bdev(dev)

        base_attr = ["cache_mode", "running", "state", "writeback_running"]
        base_path = _bcpath(dev)

        result.update(_sysfs_parse(base_path, base_attr, stats, config, internals))
        result.update(result.pop("base"))

        # ---------------- Modifications ----------------
        state = [result["state"]]
        if result.pop("running"):
            state.append("running")
        else:
            state.append("stopped")
        if "writeback_running" in result:
            if result.pop("writeback_running"):
                state.append("writeback_running")
            else:
                state.append("writeback_stopped")
        result["state"] = state

    # ---------------- Statistics ----------------
    if "stats" in result:
        replre = r"(stats|cache)_"
        statres = result["stats"]
        for attr in result["stats"]:
            if "/" not in attr:
                key = re.sub(replre, "", attr)
                statres[key] = statres.pop(attr)
            else:
                stat, key = attr.split("/", 1)
                stat = re.sub(replre, "", stat)
                key = re.sub(replre, "", key)
                if stat not in statres:
                    statres[stat] = {}
                statres[stat][key] = statres.pop(attr)
        result["stats"] = statres

    # ---------------- Internals ----------------
    if internals:
        interres = result.pop("inter_ro", {})
        interres.update(result.pop("inter_rw", {}))
        if interres:
            for key in interres:
                if key.startswith("internal"):
                    nkey = re.sub(r"internal[s/]*", "", key)
                    interres[nkey] = interres.pop(key)
                    key = nkey
                if key.startswith(("btree", "writeback")):
                    mkey, skey = re.split(r"_", key, maxsplit=1)
                    if mkey not in interres:
                        interres[mkey] = {}
                    interres[mkey][skey] = interres.pop(key)
            result["internals"] = interres

    # ---------------- Config ----------------
    if config:
        configres = result["config"]
        for key in configres:
            if key.startswith("writeback"):
                mkey, skey = re.split(r"_", key, maxsplit=1)
                if mkey not in configres:
                    configres[mkey] = {}
                configres[mkey][skey] = configres.pop(key)
        result["config"] = configres

    # ---------------- Superblock ----------------
    if superblock:
        result["superblock"] = super_(dev)

    return result


def super_(dev):
    """
    Read out BCache SuperBlock

    CLI Example:

    .. code-block:: bash

        salt '*' bcache.device bcache0
        salt '*' bcache.device /dev/sdc

    """
    dev = _devpath(dev)
    ret = {}

    res = _run_all(
        "bcache-super-show {}".format(dev),
        "error",
        "Error reading superblock on {}: %s".format(dev),
    )
    if not res:
        return False

    for line in res.splitlines():  # pylint: disable=no-member
        line = line.strip()
        if not line:
            continue

        key, val = [val.strip() for val in re.split(r"[\s]+", line, maxsplit=1)]
        if not (key and val):
            continue

        mval = None
        if " " in val:
            rval, mval = [val.strip() for val in re.split(r"[\s]+", val, maxsplit=1)]
            mval = mval[1:-1]
        else:
            rval = val

        try:
            rval = int(rval)
        except Exception:  # pylint: disable=broad-except
            try:
                rval = float(rval)
            except Exception:  # pylint: disable=broad-except
                if rval == "yes":
                    rval = True
                elif rval == "no":
                    rval = False

        pkey, key = re.split(r"\.", key, maxsplit=1)
        if pkey not in ret:
            ret[pkey] = {}

        if mval is not None:
            ret[pkey][key] = (rval, mval)
        else:
            ret[pkey][key] = rval

    return ret


# -------------------------------- HELPER FUNCTIONS --------------------------------


def _devbase(dev):
    """
    Basename of just about any dev
    """
    dev = os.path.realpath(os.path.expandvars(dev))
    dev = os.path.basename(dev)
    return dev


def _devpath(dev):
    """
    Return /dev name of just about any dev
    :return: /dev/devicename
    """
    return os.path.join("/dev", _devbase(dev))


def _syspath(dev):
    """
    Full SysFS path of a device
    """
    dev = _devbase(dev)
    dev = re.sub(r"^([vhs][a-z]+)([0-9]+)", r"\1/\1\2", dev)

    # name = re.sub(r'^([a-z]+)(?<!(bcache|md|dm))([0-9]+)', r'\1/\1\2', name)
    return os.path.join("/sys/block/", dev)


def _bdev(dev=None):
    """
    Resolve a bcacheX or cache to a real dev
    :return: basename of bcache dev
    """
    if dev is None:
        dev = _fssys("cache0")
    else:
        dev = _bcpath(dev)

    if not dev:
        return False
    else:
        return _devbase(os.path.dirname(dev))


def _bcpath(dev):
    """
    Full SysFS path of a bcache device
    """
    return os.path.join(_syspath(dev), "bcache")


def _fspath():
    """
    :return: path of active bcache
    """
    cuuid = uuid()
    if not cuuid:
        return False
    else:
        return os.path.join("/sys/fs/bcache/", cuuid)


def _fssys(name, value=None, log_lvl=None, log_msg=None):
    """
    Simple wrapper to interface with bcache SysFS
    """
    fspath = _fspath()
    if not fspath:
        return False
    else:
        return _sysfs_attr([fspath, name], value, log_lvl, log_msg)


def _bcsys(dev, name, value=None, log_lvl=None, log_msg=None):
    """
    Simple wrapper to interface with backing devs SysFS
    """
    return _sysfs_attr([_bcpath(dev), name], value, log_lvl, log_msg)


def _sysfs_attr(name, value=None, log_lvl=None, log_msg=None):
    """
    Simple wrapper with logging around sysfs.attr
    """
    if isinstance(name, str):
        name = [name]
    res = __salt__["sysfs.attr"](os.path.join(*name), value)
    if not res and log_lvl is not None and log_msg is not None:
        log.log(LOG[log_lvl], log_msg)
    return res


def _sysfs_parse(
    path, base_attr=None, stats=False, config=False, internals=False, options=False
):
    """
    Helper function for parsing BCache's SysFS interface
    """
    result = {}

    # ---------------- Parse through the interfaces list ----------------
    intfs = __salt__["sysfs.interfaces"](path)

    # Actions, we ignore
    del intfs["w"]

    # -------- Sorting hat --------
    binkeys = []
    if internals:
        binkeys.extend(["inter_ro", "inter_rw"])
    if config:
        binkeys.append("config")
    if stats:
        binkeys.append("stats")

    bintf = {}
    for key in binkeys:
        bintf[key] = []

    for intf in intfs["r"]:
        if intf.startswith("internal"):
            key = "inter_ro"
        elif "stats" in intf:
            key = "stats"
        else:
            # What to do with these???
            # I'll utilize 'inter_ro' as 'misc' as well
            key = "inter_ro"

        if key in bintf:
            bintf[key].append(intf)

    for intf in intfs["rw"]:
        if intf.startswith("internal"):
            key = "inter_rw"
        else:
            key = "config"

        if key in bintf:
            bintf[key].append(intf)

    if base_attr is not None:
        for intf in bintf:
            bintf[intf] = [sintf for sintf in bintf[intf] if sintf not in base_attr]
        bintf["base"] = base_attr

    mods = {
        "stats": [
            "internal/bset_tree_stats",
            "writeback_rate_debug",
            "metadata_written",
            "nbuckets",
            "written",
            "average_key_size",
            "btree_cache_size",
        ],
    }

    for modt, modlist in mods.items():
        found = []
        if modt not in bintf:
            continue
        for mod in modlist:
            for intflist in bintf.values():
                if mod in intflist:
                    found.append(mod)
                    intflist.remove(mod)
        bintf[modt] += found

    # -------- Fetch SysFS vals --------
    bintflist = [intf for iflist in bintf.values() for intf in iflist]
    result.update(__salt__["sysfs.read"](bintflist, path))

    # -------- Parse through well known string lists --------
    for strlist in (
        "writeback_rate_debug",
        "internal/bset_tree_stats",
        "priority_stats",
    ):
        if strlist in result:
            listres = {}
            for line in result[strlist].split("\n"):
                key, val = line.split(":", 1)
                val = val.strip()
                try:
                    val = int(val)
                except Exception:  # pylint: disable=broad-except
                    try:
                        val = float(val)
                    except Exception:  # pylint: disable=broad-except
                        pass
                listres[key.strip()] = val
            result[strlist] = listres

    # -------- Parse through selection lists --------
    if not options:
        for sellist in ("cache_mode", "cache_replacement_policy", "errors"):
            if sellist in result:
                result[sellist] = re.search(r"\[(.+)\]", result[sellist]).groups()[0]

    # -------- Parse through well known bools --------
    for boolkey in ("running", "writeback_running", "congested"):
        if boolkey in result:
            result[boolkey] = bool(result[boolkey])

    # -------- Recategorize results --------
    bresult = {}
    for iftype, intflist in bintf.items():
        ifres = {}
        for intf in intflist:
            if intf in result:
                ifres[intf] = result.pop(intf)
        if ifres:
            bresult[iftype] = ifres

    return bresult


def _size_map(size):
    """
    Map Bcache's size strings to real bytes
    """
    try:
        # I know, I know, EAFP.
        # But everything else is reason for None
        if not isinstance(size, int):
            if re.search(r"[Kk]", size):
                size = 1024 * float(re.sub(r"[Kk]", "", size))
            elif re.search(r"[Mm]", size):
                size = 1024 ** 2 * float(re.sub(r"[Mm]", "", size))
            size = int(size)
        return size
    except Exception:  # pylint: disable=broad-except
        return None


def _sizes(dev):
    """
    Return neigh useless sizing info about a blockdev
    :return: (total size in blocks, blocksize, maximum discard size in bytes)
    """
    dev = _devbase(dev)

    # standarization yay
    block_sizes = (
        "hw_sector_size",
        "minimum_io_size",
        "physical_block_size",
        "logical_block_size",
    )
    discard_sizes = (
        "discard_max_bytes",
        "discard_max_hw_bytes",
    )

    sysfs = __salt__["sysfs.read"](
        (
            "size",
            "queue/hw_sector_size",
            "../queue/hw_sector_size",
            "queue/discard_max_bytes",
            "../queue/discard_max_bytes",
        ),
        root=_syspath(dev),
    )

    # TODO: makes no sense
    # First of all, it has to be a power of 2
    # Secondly, this returns 4GiB - 512b on Intel 3500's for some weird reason
    # discard_granularity seems in bytes, resolves to 512b ???
    # max_hw_sectors_kb???
    # There's also discard_max_hw_bytes more recently
    # See: https://www.kernel.org/doc/Documentation/block/queue-sysfs.txt
    # Also, I cant find any docs yet regarding bucket sizes;
    # it's supposed to be discard_max_hw_bytes,
    # but no way to figure that one reliably out apparently

    discard = sysfs.get(
        "queue/discard_max_bytes", sysfs.get("../queue/discard_max_bytes", None)
    )
    block = sysfs.get(
        "queue/hw_sector_size", sysfs.get("../queue/hw_sector_size", None)
    )

    return 512 * sysfs["size"], block, discard


def _wipe(dev):
    """
    REALLY DESTRUCTIVE STUFF RIGHT AHEAD
    """
    endres = 0
    dev = _devbase(dev)

    size, block, discard = _sizes(dev)

    if discard is None:
        log.error("Unable to read SysFS props for %s", dev)
        return None
    elif not discard:
        log.warning("%s seems unable to discard", dev)
        wiper = "dd"
    elif not HAS_BLKDISCARD:
        log.warning(
            "blkdiscard binary not available, properly wipe the dev manually for"
            " optimal results"
        )
        wiper = "dd"
    else:
        wiper = "blkdiscard"

    wipe_failmsg = "Error wiping {}: %s".format(dev)
    if wiper == "dd":
        blocks = 4
        cmd = "dd if=/dev/zero of=/dev/{} bs=1M count={}".format(dev, blocks)
        endres += _run_all(cmd, "warn", wipe_failmsg)

        # Some stuff (<cough>GPT</cough>) writes stuff at the end of a dev as well
        cmd += " seek={}".format((size / 1024 ** 2) - blocks)
        endres += _run_all(cmd, "warn", wipe_failmsg)

    elif wiper == "blkdiscard":
        cmd = "blkdiscard /dev/{}".format(dev)
        endres += _run_all(cmd, "warn", wipe_failmsg)
        # TODO: fix annoying bug failing blkdiscard by trying to discard 1 sector past blkdev
        endres = 1

    return endres > 0


def _wait(lfunc, log_lvl=None, log_msg=None, tries=10):
    """
    Wait for lfunc to be True
    :return: True if lfunc succeeded within tries, False if it didn't
    """
    i = 0
    while i < tries:
        time.sleep(1)

        if lfunc():
            return True
        else:
            i += 1
    if log_lvl is not None:
        log.log(LOG[log_lvl], log_msg)
    return False


def _run_all(cmd, log_lvl=None, log_msg=None, exitcode=0):
    """
    Simple wrapper around cmd.run_all
    log_msg can contain {0} for stderr
    :return: True or stdout, False if retcode wasn't exitcode
    """
    res = __salt__["cmd.run_all"](cmd)
    if res["retcode"] == exitcode:
        if res["stdout"]:
            return res["stdout"]
        else:
            return True

    if log_lvl is not None:
        log.log(LOG[log_lvl], log_msg, res["stderr"])
    return False


def _alltrue(resdict):
    if resdict is None:
        return True
    return len([val for val in resdict.values() if val]) > 0



"""
Azure (ARM) Resource Execution Module

.. versionadded:: 2019.2.0

:maintainer: <devops@decisionlab.io>
:maturity: new
:depends:
    * `azure <https://pypi.python.org/pypi/azure>`_ >= 2.0.0
    * `azure-common <https://pypi.python.org/pypi/azure-common>`_ >= 1.1.8
    * `azure-mgmt <https://pypi.python.org/pypi/azure-mgmt>`_ >= 1.0.0
    * `azure-mgmt-compute <https://pypi.python.org/pypi/azure-mgmt-compute>`_ >= 1.0.0
    * `azure-mgmt-network <https://pypi.python.org/pypi/azure-mgmt-network>`_ >= 1.7.1
    * `azure-mgmt-resource <https://pypi.python.org/pypi/azure-mgmt-resource>`_ >= 1.1.0
    * `azure-mgmt-storage <https://pypi.python.org/pypi/azure-mgmt-storage>`_ >= 1.0.0
    * `azure-mgmt-web <https://pypi.python.org/pypi/azure-mgmt-web>`_ >= 0.32.0
    * `azure-storage <https://pypi.python.org/pypi/azure-storage>`_ >= 0.34.3
    * `msrestazure <https://pypi.python.org/pypi/msrestazure>`_ >= 0.4.21
:platform: linux

:configuration: This module requires Azure Resource Manager credentials to be passed as keyword arguments
    to every function in order to work properly.

    Required provider parameters:

    if using username and password:
      * ``subscription_id``
      * ``username``
      * ``password``

    if using a service principal:
      * ``subscription_id``
      * ``tenant``
      * ``client_id``
      * ``secret``

    Optional provider parameters:

**cloud_environment**: Used to point the cloud driver to different API endpoints, such as Azure GovCloud.
    Possible values:
      * ``AZURE_PUBLIC_CLOUD`` (default)
      * ``AZURE_CHINA_CLOUD``
      * ``AZURE_US_GOV_CLOUD``
      * ``AZURE_GERMAN_CLOUD``

"""

# Python libs

import logging

# Salt Libs
import salt.utils.json

# Azure libs
HAS_LIBS = False
try:
    import azure.mgmt.resource.resources.models  # pylint: disable=unused-import
    from msrest.exceptions import SerializationError
    from msrestazure.azure_exceptions import CloudError

    HAS_LIBS = True
except ImportError:
    pass

__virtualname__ = "azurearm_resource"

log = logging.getLogger(__name__)


def __virtual__():
    if not HAS_LIBS:
        return (
            False,
            "The following dependencies are required to use the AzureARM modules: "
            "Microsoft Azure SDK for Python >= 2.0rc6, "
            "MS REST Azure (msrestazure) >= 0.4",
        )

    return __virtualname__


def resource_groups_list(**kwargs):
    """
    .. versionadded:: 2019.2.0

    List all resource groups within a subscription.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_resource.resource_groups_list

    """
    result = {}
    resconn = __utils__["azurearm.get_client"]("resource", **kwargs)
    try:
        groups = __utils__["azurearm.paged_object_to_list"](
            resconn.resource_groups.list()
        )

        for group in groups:
            result[group["name"]] = group
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("resource", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


def resource_group_check_existence(name, **kwargs):
    """
    .. versionadded:: 2019.2.0

    Check for the existence of a named resource group in the current subscription.

    :param name: The resource group name to check.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_resource.resource_group_check_existence testgroup

    """
    result = False
    resconn = __utils__["azurearm.get_client"]("resource", **kwargs)
    try:
        result = resconn.resource_groups.check_existence(name)

    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("resource", str(exc), **kwargs)

    return result


def resource_group_get(name, **kwargs):
    """
    .. versionadded:: 2019.2.0

    Get a dictionary representing a resource group's properties.

    :param name: The resource group name to get.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_resource.resource_group_get testgroup

    """
    result = {}
    resconn = __utils__["azurearm.get_client"]("resource", **kwargs)
    try:
        group = resconn.resource_groups.get(name)
        result = group.as_dict()

    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("resource", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


def resource_group_create_or_update(
    name, location, **kwargs
):  # pylint: disable=invalid-name
    """
    .. versionadded:: 2019.2.0

    Create or update a resource group in a given location.

    :param name: The name of the resource group to create or update.

    :param location: The location of the resource group. This value
        is not able to be updated once the resource group is created.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_resource.resource_group_create_or_update testgroup westus

    """
    result = {}
    resconn = __utils__["azurearm.get_client"]("resource", **kwargs)
    resource_group_params = {
        "location": location,
        "managed_by": kwargs.get("managed_by"),
        "tags": kwargs.get("tags"),
    }
    try:
        group = resconn.resource_groups.create_or_update(name, resource_group_params)
        result = group.as_dict()
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("resource", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


def resource_group_delete(name, **kwargs):
    """
    .. versionadded:: 2019.2.0

    Delete a resource group from the subscription.

    :param name: The resource group name to delete.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_resource.resource_group_delete testgroup

    """
    result = False
    resconn = __utils__["azurearm.get_client"]("resource", **kwargs)
    try:
        group = resconn.resource_groups.delete(name)
        group.wait()
        result = True
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("resource", str(exc), **kwargs)

    return result


def deployment_operation_get(operation, deployment, resource_group, **kwargs):
    """
    .. versionadded:: 2019.2.0

    Get a deployment operation within a deployment.

    :param operation: The operation ID of the operation within the deployment.

    :param deployment: The name of the deployment containing the operation.

    :param resource_group: The resource group name assigned to the
        deployment.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_resource.deployment_operation_get XXXXX testdeploy testgroup

    """
    resconn = __utils__["azurearm.get_client"]("resource", **kwargs)
    try:
        operation = resconn.deployment_operations.get(
            resource_group_name=resource_group,
            deployment_name=deployment,
            operation_id=operation,
        )

        result = operation.as_dict()
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("resource", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


def deployment_operations_list(name, resource_group, result_limit=10, **kwargs):
    """
    .. versionadded:: 2019.2.0

    List all deployment operations within a deployment.

    :param name: The name of the deployment to query.

    :param resource_group: The resource group name assigned to the
        deployment.

    :param result_limit: (Default: 10) The limit on the list of deployment
        operations.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_resource.deployment_operations_list testdeploy testgroup

    """
    result = {}
    resconn = __utils__["azurearm.get_client"]("resource", **kwargs)
    try:
        operations = __utils__["azurearm.paged_object_to_list"](
            resconn.deployment_operations.list(
                resource_group_name=resource_group,
                deployment_name=name,
                top=result_limit,
            )
        )

        for oper in operations:
            result[oper["operation_id"]] = oper
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("resource", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


def deployment_delete(name, resource_group, **kwargs):
    """
    .. versionadded:: 2019.2.0

    Delete a deployment.

    :param name: The name of the deployment to delete.

    :param resource_group: The resource group name assigned to the
        deployment.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_resource.deployment_delete testdeploy testgroup

    """
    result = False
    resconn = __utils__["azurearm.get_client"]("resource", **kwargs)
    try:
        deploy = resconn.deployments.delete(
            deployment_name=name, resource_group_name=resource_group
        )
        deploy.wait()
        result = True
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("resource", str(exc), **kwargs)

    return result


def deployment_check_existence(name, resource_group, **kwargs):
    """
    .. versionadded:: 2019.2.0

    Check the existence of a deployment.

    :param name: The name of the deployment to query.

    :param resource_group: The resource group name assigned to the
        deployment.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_resource.deployment_check_existence testdeploy testgroup

    """
    result = False
    resconn = __utils__["azurearm.get_client"]("resource", **kwargs)
    try:
        result = resconn.deployments.check_existence(
            deployment_name=name, resource_group_name=resource_group
        )
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("resource", str(exc), **kwargs)

    return result


def deployment_create_or_update(
    name,
    resource_group,
    deploy_mode="incremental",
    debug_setting="none",
    deploy_params=None,
    parameters_link=None,
    deploy_template=None,
    template_link=None,
    **kwargs
):
    """
    .. versionadded:: 2019.2.0

    Deploys resources to a resource group.

    :param name: The name of the deployment to create or update.

    :param resource_group: The resource group name assigned to the
        deployment.

    :param deploy_mode: The mode that is used to deploy resources. This value can be either
        'incremental' or 'complete'. In Incremental mode, resources are deployed without deleting
        existing resources that are not included in the template. In Complete mode, resources
        are deployed and existing resources in the resource group that are not included in
        the template are deleted. Be careful when using Complete mode as you may
        unintentionally delete resources.

    :param debug_setting: The debug setting of the deployment. The permitted values are 'none',
        'requestContent', 'responseContent', or 'requestContent,responseContent'. By logging
        information about the request or response, you could potentially expose sensitive data
        that is retrieved through the deployment operations.

    :param deploy_params: JSON string containing name and value pairs that define the deployment
        parameters for the template. You use this element when you want to provide the parameter
        values directly in the request rather than link to an existing parameter file. Use either
        the parameters_link property or the deploy_params property, but not both.

    :param parameters_link: The URI of a parameters file. You use this element to link to an existing
        parameters file. Use either the parameters_link property or the deploy_params property, but not both.

    :param deploy_template: JSON string of template content. You use this element when you want to pass
        the template syntax directly in the request rather than link to an existing template. Use either
        the template_link property or the deploy_template property, but not both.

    :param template_link: The URI of the template. Use either the template_link property or the
        deploy_template property, but not both.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_resource.deployment_create_or_update testdeploy testgroup

    """
    resconn = __utils__["azurearm.get_client"]("resource", **kwargs)

    prop_kwargs = {"mode": deploy_mode}
    prop_kwargs["debug_setting"] = {"detail_level": debug_setting}

    if deploy_params:
        prop_kwargs["parameters"] = deploy_params
    else:
        if isinstance(parameters_link, dict):
            prop_kwargs["parameters_link"] = parameters_link
        else:
            prop_kwargs["parameters_link"] = {"uri": parameters_link}

    if deploy_template:
        prop_kwargs["template"] = deploy_template
    else:
        if isinstance(template_link, dict):
            prop_kwargs["template_link"] = template_link
        else:
            prop_kwargs["template_link"] = {"uri": template_link}

    deploy_kwargs = kwargs.copy()
    deploy_kwargs.update(prop_kwargs)

    try:
        deploy_model = __utils__["azurearm.create_object_model"](
            "resource", "DeploymentProperties", **deploy_kwargs
        )
    except TypeError as exc:
        result = {"error": "The object model could not be built. ({})".format(str(exc))}
        return result

    try:
        validate = deployment_validate(
            name=name, resource_group=resource_group, **deploy_kwargs
        )
        if "error" in validate:
            result = validate
        else:
            deploy = resconn.deployments.create_or_update(
                deployment_name=name,
                resource_group_name=resource_group,
                properties=deploy_model,
            )
            deploy.wait()
            deploy_result = deploy.result()
            result = deploy_result.as_dict()
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("resource", str(exc), **kwargs)
        result = {"error": str(exc)}
    except SerializationError as exc:
        result = {
            "error": "The object model could not be parsed. ({})".format(str(exc))
        }

    return result


def deployment_get(name, resource_group, **kwargs):
    """
    .. versionadded:: 2019.2.0

    Get details about a specific deployment.

    :param name: The name of the deployment to query.

    :param resource_group: The resource group name assigned to the
        deployment.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_resource.deployment_get testdeploy testgroup

    """
    resconn = __utils__["azurearm.get_client"]("resource", **kwargs)
    try:
        deploy = resconn.deployments.get(
            deployment_name=name, resource_group_name=resource_group
        )
        result = deploy.as_dict()
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("resource", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


def deployment_cancel(name, resource_group, **kwargs):
    """
    .. versionadded:: 2019.2.0

    Cancel a deployment if in 'Accepted' or 'Running' state.

    :param name: The name of the deployment to cancel.

    :param resource_group: The resource group name assigned to the
        deployment.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_resource.deployment_cancel testdeploy testgroup

    """
    resconn = __utils__["azurearm.get_client"]("resource", **kwargs)
    try:
        resconn.deployments.cancel(
            deployment_name=name, resource_group_name=resource_group
        )
        result = {"result": True}
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("resource", str(exc), **kwargs)
        result = {"error": str(exc), "result": False}

    return result


def deployment_validate(
    name,
    resource_group,
    deploy_mode=None,
    debug_setting=None,
    deploy_params=None,
    parameters_link=None,
    deploy_template=None,
    template_link=None,
    **kwargs
):
    """
    .. versionadded:: 2019.2.0

    Validates whether the specified template is syntactically correct
    and will be accepted by Azure Resource Manager.

    :param name: The name of the deployment to validate.

    :param resource_group: The resource group name assigned to the
        deployment.

    :param deploy_mode: The mode that is used to deploy resources. This value can be either
        'incremental' or 'complete'. In Incremental mode, resources are deployed without deleting
        existing resources that are not included in the template. In Complete mode, resources
        are deployed and existing resources in the resource group that are not included in
        the template are deleted. Be careful when using Complete mode as you may
        unintentionally delete resources.

    :param debug_setting: The debug setting of the deployment. The permitted values are 'none',
        'requestContent', 'responseContent', or 'requestContent,responseContent'. By logging
        information about the request or response, you could potentially expose sensitive data
        that is retrieved through the deployment operations.

    :param deploy_params: JSON string containing name and value pairs that define the deployment
        parameters for the template. You use this element when you want to provide the parameter
        values directly in the request rather than link to an existing parameter file. Use either
        the parameters_link property or the deploy_params property, but not both.

    :param parameters_link: The URI of a parameters file. You use this element to link to an existing
        parameters file. Use either the parameters_link property or the deploy_params property, but not both.

    :param deploy_template: JSON string of template content. You use this element when you want to pass
        the template syntax directly in the request rather than link to an existing template. Use either
        the template_link property or the deploy_template property, but not both.

    :param template_link: The URI of the template. Use either the template_link property or the
        deploy_template property, but not both.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_resource.deployment_validate testdeploy testgroup

    """
    resconn = __utils__["azurearm.get_client"]("resource", **kwargs)

    prop_kwargs = {"mode": deploy_mode}
    prop_kwargs["debug_setting"] = {"detail_level": debug_setting}

    if deploy_params:
        prop_kwargs["parameters"] = deploy_params
    else:
        if isinstance(parameters_link, dict):
            prop_kwargs["parameters_link"] = parameters_link
        else:
            prop_kwargs["parameters_link"] = {"uri": parameters_link}

    if deploy_template:
        prop_kwargs["template"] = deploy_template
    else:
        if isinstance(template_link, dict):
            prop_kwargs["template_link"] = template_link
        else:
            prop_kwargs["template_link"] = {"uri": template_link}

    deploy_kwargs = kwargs.copy()
    deploy_kwargs.update(prop_kwargs)

    try:
        deploy_model = __utils__["azurearm.create_object_model"](
            "resource", "DeploymentProperties", **deploy_kwargs
        )
    except TypeError as exc:
        result = {"error": "The object model could not be built. ({})".format(str(exc))}
        return result

    try:
        local_validation = deploy_model.validate()
        if local_validation:
            raise local_validation[0]

        deploy = resconn.deployments.validate(
            deployment_name=name,
            resource_group_name=resource_group,
            properties=deploy_model,
        )
        result = deploy.as_dict()
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("resource", str(exc), **kwargs)
        result = {"error": str(exc)}
    except SerializationError as exc:
        result = {
            "error": "The object model could not be parsed. ({})".format(str(exc))
        }

    return result


def deployment_export_template(name, resource_group, **kwargs):
    """
    .. versionadded:: 2019.2.0

    Exports the template used for the specified deployment.

    :param name: The name of the deployment to query.

    :param resource_group: The resource group name assigned to the
        deployment.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_resource.deployment_export_template testdeploy testgroup

    """
    resconn = __utils__["azurearm.get_client"]("resource", **kwargs)
    try:
        deploy = resconn.deployments.export_template(
            deployment_name=name, resource_group_name=resource_group
        )
        result = deploy.as_dict()
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("resource", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


def deployments_list(resource_group, **kwargs):
    """
    .. versionadded:: 2019.2.0

    List all deployments within a resource group.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_resource.deployments_list testgroup

    """
    result = {}
    resconn = __utils__["azurearm.get_client"]("resource", **kwargs)
    try:
        deployments = __utils__["azurearm.paged_object_to_list"](
            resconn.deployments.list_by_resource_group(
                resource_group_name=resource_group
            )
        )

        for deploy in deployments:
            result[deploy["name"]] = deploy
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("resource", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


def subscriptions_list_locations(subscription_id=None, **kwargs):
    """
    .. versionadded:: 2019.2.0

    List all locations for a subscription.

    :param subscription_id: The ID of the subscription to query.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_resource.subscriptions_list_locations XXXXXXXX

    """
    result = {}

    if not subscription_id:
        subscription_id = kwargs.get("subscription_id")
    elif not kwargs.get("subscription_id"):
        kwargs["subscription_id"] = subscription_id

    subconn = __utils__["azurearm.get_client"]("subscription", **kwargs)
    try:
        locations = __utils__["azurearm.paged_object_to_list"](
            subconn.subscriptions.list_locations(
                subscription_id=kwargs["subscription_id"]
            )
        )

        for loc in locations:
            result[loc["name"]] = loc
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("resource", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


def subscription_get(subscription_id=None, **kwargs):
    """
    .. versionadded:: 2019.2.0

    Get details about a subscription.

    :param subscription_id: The ID of the subscription to query.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_resource.subscription_get XXXXXXXX

    """
    result = {}

    if not subscription_id:
        subscription_id = kwargs.get("subscription_id")
    elif not kwargs.get("subscription_id"):
        kwargs["subscription_id"] = subscription_id

    subconn = __utils__["azurearm.get_client"]("subscription", **kwargs)
    try:
        subscription = subconn.subscriptions.get(
            subscription_id=kwargs.get("subscription_id")
        )

        result = subscription.as_dict()
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("resource", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


def subscriptions_list(**kwargs):
    """
    .. versionadded:: 2019.2.0

    List all subscriptions for a tenant.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_resource.subscriptions_list

    """
    result = {}
    subconn = __utils__["azurearm.get_client"]("subscription", **kwargs)
    try:
        subs = __utils__["azurearm.paged_object_to_list"](subconn.subscriptions.list())

        for sub in subs:
            result[sub["subscription_id"]] = sub
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("resource", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


def tenants_list(**kwargs):
    """
    .. versionadded:: 2019.2.0

    List all tenants for your account.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_resource.tenants_list

    """
    result = {}
    subconn = __utils__["azurearm.get_client"]("subscription", **kwargs)
    try:
        tenants = __utils__["azurearm.paged_object_to_list"](subconn.tenants.list())

        for tenant in tenants:
            result[tenant["tenant_id"]] = tenant
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("resource", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


def policy_assignment_delete(name, scope, **kwargs):
    """
    .. versionadded:: 2019.2.0

    Delete a policy assignment.

    :param name: The name of the policy assignment to delete.

    :param scope: The scope of the policy assignment.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_resource.policy_assignment_delete testassign \
        /subscriptions/bc75htn-a0fhsi-349b-56gh-4fghti-f84852

    """
    result = False
    polconn = __utils__["azurearm.get_client"]("policy", **kwargs)
    try:
        # pylint: disable=unused-variable
        policy = polconn.policy_assignments.delete(
            policy_assignment_name=name, scope=scope
        )
        result = True
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("resource", str(exc), **kwargs)

    return result


def policy_assignment_create(name, scope, definition_name, **kwargs):
    """
    .. versionadded:: 2019.2.0

    Create a policy assignment.

    :param name: The name of the policy assignment to create.

    :param scope: The scope of the policy assignment.

    :param definition_name: The name of the policy definition to assign.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_resource.policy_assignment_create testassign \
        /subscriptions/bc75htn-a0fhsi-349b-56gh-4fghti-f84852 testpolicy

    """
    polconn = __utils__["azurearm.get_client"]("policy", **kwargs)

    # "get" doesn't work for built-in policies per https://github.com/Azure/azure-cli/issues/692
    # Uncomment this section when the ticket above is resolved.
    #  BEGIN
    # definition = policy_definition_get(
    #     name=definition_name,
    #     **kwargs
    # )
    #  END

    # Delete this section when the ticket above is resolved.
    #  BEGIN
    definition_list = policy_definitions_list(**kwargs)
    if definition_name in definition_list:
        definition = definition_list[definition_name]
    else:
        definition = {
            "error": 'The policy definition named "{}" could not be found.'.format(
                definition_name
            )
        }
    #  END

    if "error" not in definition:
        definition_id = str(definition["id"])

        prop_kwargs = {"policy_definition_id": definition_id}

        policy_kwargs = kwargs.copy()
        policy_kwargs.update(prop_kwargs)

        try:
            policy_model = __utils__["azurearm.create_object_model"](
                "resource.policy", "PolicyAssignment", **policy_kwargs
            )
        except TypeError as exc:
            result = {
                "error": "The object model could not be built. ({})".format(str(exc))
            }
            return result

        try:
            policy = polconn.policy_assignments.create(
                scope=scope, policy_assignment_name=name, parameters=policy_model
            )
            result = policy.as_dict()
        except CloudError as exc:
            __utils__["azurearm.log_cloud_error"]("resource", str(exc), **kwargs)
            result = {"error": str(exc)}
        except SerializationError as exc:
            result = {
                "error": "The object model could not be parsed. ({})".format(str(exc))
            }
    else:
        result = {
            "error": 'The policy definition named "{}" could not be found.'.format(
                definition_name
            )
        }

    return result


def policy_assignment_get(name, scope, **kwargs):
    """
    .. versionadded:: 2019.2.0

    Get details about a specific policy assignment.

    :param name: The name of the policy assignment to query.

    :param scope: The scope of the policy assignment.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_resource.policy_assignment_get testassign \
        /subscriptions/bc75htn-a0fhsi-349b-56gh-4fghti-f84852

    """
    polconn = __utils__["azurearm.get_client"]("policy", **kwargs)
    try:
        policy = polconn.policy_assignments.get(
            policy_assignment_name=name, scope=scope
        )
        result = policy.as_dict()
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("resource", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


def policy_assignments_list_for_resource_group(
    resource_group, **kwargs
):  # pylint: disable=invalid-name
    """
    .. versionadded:: 2019.2.0

    List all policy assignments for a resource group.

    :param resource_group: The resource group name to list policy assignments within.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_resource.policy_assignments_list_for_resource_group testgroup

    """
    result = {}
    polconn = __utils__["azurearm.get_client"]("policy", **kwargs)
    try:
        policy_assign = __utils__["azurearm.paged_object_to_list"](
            polconn.policy_assignments.list_for_resource_group(
                resource_group_name=resource_group, filter=kwargs.get("filter")
            )
        )

        for assign in policy_assign:
            result[assign["name"]] = assign
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("resource", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


def policy_assignments_list(**kwargs):
    """
    .. versionadded:: 2019.2.0

    List all policy assignments for a subscription.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_resource.policy_assignments_list

    """
    result = {}
    polconn = __utils__["azurearm.get_client"]("policy", **kwargs)
    try:
        policy_assign = __utils__["azurearm.paged_object_to_list"](
            polconn.policy_assignments.list()
        )

        for assign in policy_assign:
            result[assign["name"]] = assign
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("resource", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


def policy_definition_create_or_update(
    name, policy_rule, **kwargs
):  # pylint: disable=invalid-name
    """
    .. versionadded:: 2019.2.0

    Create or update a policy definition.

    :param name: The name of the policy definition to create or update.

    :param policy_rule: A dictionary defining the
        `policy rule <https://docs.microsoft.com/en-us/azure/azure-policy/policy-definition#policy-rule>`_.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_resource.policy_definition_create_or_update testpolicy '{...rule definition..}'

    """
    if not isinstance(policy_rule, dict):
        result = {"error": "The policy rule must be a dictionary!"}
        return result

    polconn = __utils__["azurearm.get_client"]("policy", **kwargs)

    # Convert OrderedDict to dict
    prop_kwargs = {
        "policy_rule": salt.utils.json.loads(salt.utils.json.dumps(policy_rule))
    }

    policy_kwargs = kwargs.copy()
    policy_kwargs.update(prop_kwargs)

    try:
        policy_model = __utils__["azurearm.create_object_model"](
            "resource.policy", "PolicyDefinition", **policy_kwargs
        )
    except TypeError as exc:
        result = {"error": "The object model could not be built. ({})".format(str(exc))}
        return result

    try:
        policy = polconn.policy_definitions.create_or_update(
            policy_definition_name=name, parameters=policy_model
        )
        result = policy.as_dict()
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("resource", str(exc), **kwargs)
        result = {"error": str(exc)}
    except SerializationError as exc:
        result = {
            "error": "The object model could not be parsed. ({})".format(str(exc))
        }

    return result


def policy_definition_delete(name, **kwargs):
    """
    .. versionadded:: 2019.2.0

    Delete a policy definition.

    :param name: The name of the policy definition to delete.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_resource.policy_definition_delete testpolicy

    """
    result = False
    polconn = __utils__["azurearm.get_client"]("policy", **kwargs)
    try:
        # pylint: disable=unused-variable
        policy = polconn.policy_definitions.delete(policy_definition_name=name)
        result = True
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("resource", str(exc), **kwargs)

    return result


def policy_definition_get(name, **kwargs):
    """
    .. versionadded:: 2019.2.0

    Get details about a specific policy definition.

    :param name: The name of the policy definition to query.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_resource.policy_definition_get testpolicy

    """
    polconn = __utils__["azurearm.get_client"]("policy", **kwargs)
    try:
        policy_def = polconn.policy_definitions.get(policy_definition_name=name)
        result = policy_def.as_dict()
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("resource", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


def policy_definitions_list(hide_builtin=False, **kwargs):
    """
    .. versionadded:: 2019.2.0

    List all policy definitions for a subscription.

    :param hide_builtin: Boolean which will filter out BuiltIn policy definitions from the result.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_resource.policy_definitions_list

    """
    result = {}
    polconn = __utils__["azurearm.get_client"]("policy", **kwargs)
    try:
        policy_defs = __utils__["azurearm.paged_object_to_list"](
            polconn.policy_definitions.list()
        )

        for policy in policy_defs:
            if not (hide_builtin and policy["policy_type"] == "BuiltIn"):
                result[policy["name"]] = policy
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("resource", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result
